within OpenWPL;

package Functions "External functions"
  extends Modelica.Icons.FunctionsPackage;
  annotation (
    version="1.0.0",
    versionDate="08.06.21",
    Protection(access = Access.packageDuplicate),
    preferredView="info",
    Documentation(info="<html>
<p> A package of external functions which is used in the library. Most functions are based on the method of characteristics. </p>
</html>"),
  Icon);
  
  function Re
    extends Modelica.Icons.Function;
    input Modelica.SIunits.Velocity v "Flow velocity";
    input Modelica.SIunits.Diameter D "Pipe diameter";
    input Modelica.SIunits.Density rho "Density";
    input Modelica.SIunits.DynamicViscosity mu "Dynamic viscosity of water";
    output Modelica.SIunits.ReynoldsNumber Re_D "Reynolds number";
  algorithm
    Re_D := (rho*abs(v)*D)/mu;
  end Re;
  
  function f_Haaland
    extends Modelica.Icons.Function;
    input Modelica.SIunits.Velocity v "Flow velocity";
    input Modelica.SIunits.Diameter D "Pipe diameter";
    input Modelica.SIunits.Density rho "Density";
    input Modelica.SIunits.DynamicViscosity mu "Dynamic viscosity of water";
    input Modelica.SIunits.Height eps "Pipe roughness height";
    output Real f "Friction factor";
  protected
    Modelica.SIunits.ReynoldsNumber Re_D "Reynolds number";
  algorithm
    Re_D := Re(v, D, rho, mu);
    f :=  100*(1/(-1.8*log(((eps/D)/3.7)^(1.11)+(6.9/Re_D))))^2;
  end f_Haaland;

  function B_Impedance
    extends Modelica.Icons.Function;
    input Modelica.SIunits.Acceleration g "Graviational accelleration";
    input Modelica.SIunits.VelocityOfSound a "Velocity of sound in water";
    input Modelica.SIunits.Area A "Area";
    output Real B;
  algorithm
    B := a/(g*A);
  end B_Impedance;

  function R_Resistance
    extends Modelica.Icons.Function;
    input Real f "Friction factor";
    input Modelica.SIunits.Acceleration g "Graviational accelleration";
    input Modelica.SIunits.Diameter D "Pipe diameter";
    input Modelica.SIunits.Area A "Pipe area";
    input Modelica.SIunits.Distance dx;
    output Real R;
  algorithm
    R := f * dx / (2 * g * D * A ^ 2);
  end R_Resistance;

  function N_parts
    extends Modelica.Icons.Function;
    
    input Modelica.SIunits.Length L;
    input Modelica.SIunits.Length dx;
    output Real N;
  algorithm
    N := L/dx;
  end N_parts;

  function T_step
    extends Modelica.Icons.Function;
  
    input Modelica.SIunits.Length dx;
    input Modelica.SIunits.Velocity a;
    output Modelica.SIunits.Time dt;
  algorithm
    dt := dx/a;
  end T_step;

  function C_plus
    extends Modelica.Icons.Function;
    
    /*
    input Real Hb;
    input Real B;
    input Real Qb;
    input Real R;
    output Real Cp;
    */
    
    input Real Ha "Hydraulic head at point A";
    input Real B "Impedance";
    input Real Qa "Volumetric flow at point A";
    output Real Cp "Plus characteristic given in equation (3-23) in Wylie & Streeter";
  algorithm
//Cp := Hb - B*Qb + R*Qb*abs(Qb);
    Cp := Ha + B*Qa;
  end C_plus;

  function C_minus
    extends Modelica.Icons.Function;
    
    /*
    input Real Ha;
    input Real B;
    input Real Qa;
    input Real R;
    output Real Cm;
    */
    
    input Real Hb "Hydraulic head at point B";
    input Real B "Impedance";
    input Real Qb "Volumetric flow at point B";
    output Real Cm "Minus characteristic given in equation (3-24) in Wylie & Streeter";
  algorithm
//Cm := Ha + B*Qa - R*Qa*abs(Qa);
    Cm := Hb - B*Qb;
  end C_minus;

  function H_head
    extends Modelica.Icons.Function;
    input Real Cp "Plus characteristic given in equation (3-24) in Wylie & Streeter";
    input Real Cm "Minus characteristic given in equation (3-24) in Wylie & Streeter";
    input Real Bm "Impedance at Cm given in equation (3-24) in Wylie & Streeter";
    input Real Bp "Impedance at Cp given in equation (3-23) in Wylie & Streeter";
    output Real H "Hydraulic head given in equation (2-25) in Wylie & Streeter";
  algorithm
    H := (Cp*Bm+Cm*Bp)/(Bp+Bm);
  end H_head;

  function Q_flow
    extends Modelica.Icons.Function;
    input Real Cp "Plus characteristic given in equation (3-24) in Wylie & Streeter";
    input Real Cm "Minus characteristic given in equation (3-24) in Wylie & Streeter";
    input Real Bm "Impedance at Cm given in equation (3-24) in Wylie & Streeter";
    input Real Bp "Impedance at Cp given in equation (3-23) in Wylie & Streeter";
    output Real Q "Volumetric flow given in equation (2-26) in Wylie & Streeter";
  algorithm
    Q := (Cp-Cm)/(Bp+Bm);
  end Q_flow;

  function v_Velocity
    extends Modelica.Icons.Function;
    input Real H "Hydraulic head";
    input Modelica.SIunits.Acceleration g "Graviational accelleration";
    output Modelica.SIunits.Velocity v "Flow velocity";
  algorithm
    v := sqrt(2*g*abs(H));
  end v_Velocity;

  function H_expansion
    extends Modelica.Icons.Function;
    input Real Ha;
    input Modelica.SIunits.Acceleration g "Graviational accelleration";
    input Modelica.SIunits.VelocityOfSound a "Velocity of sound in water";
    input Modelica.SIunits.Area Aa "Area at point A";
    input Modelica.SIunits.Area A "Area at point P";
    input Real Qa "Volumetric flow at point A";
    input Real Q "Volumetric flow at point P";
    output Real H "Hydraulic head";
  algorithm
    H := Ha - (a/g)*(Q/A-Qa/Aa-((Q+Qa)/2)*(1/A - 1/Aa));
  end H_expansion;

  function Q_expansion
    extends Modelica.Icons.Function;
    input Real Ha;
    input Real Hb;
    input Modelica.SIunits.Acceleration g "Graviational accelleration";
    input Modelica.SIunits.VelocityOfSound a "Velocity of sound in water";
    input Modelica.SIunits.Area Ab "Area at point B";
    input Modelica.SIunits.Area Aa "Area at point A";
    input Modelica.SIunits.Area A "Area at point P";
    input Real Qa "Volumetric flow at point A";
    input Real Qb "Volumetric flow at point B";
    output Real Q "Hydraulic head";
  algorithm
  Q := ((g/a)*(Ha-Hb) + Qb*(1/(2*A) + 1/(2*Ab)) + Qa*(1/(2*A) + 1/(2*Aa)))/(1/A + 1/(2*Ab) + 1/(2*Aa));
  end Q_expansion;

  function H_initial
  extends Modelica.Icons.Function;
    input Real f "Friction factor";
    input Real H_res "Head at reservoir";
    input Real Q "Volumetric flow rate";
    input Modelica.SIunits.Length L;
    input Modelica.SIunits.Acceleration g "Graviational accelleration";
    input Modelica.SIunits.Diameter D "Pipe diameter";
    input Modelica.SIunits.Area A "Pipe area";
    output Real H;
  algorithm
    H := H_res - f*(L/D)*abs(Q)*Q/(2*g*A^2);
  end H_initial;

  function v_Bernoulli
    extends Modelica.Icons.Function;
    input Modelica.SIunits.Height H_in "H. head at reservoir";
    input Modelica.SIunits.Height H_out "H. head at sink";
    input Modelica.SIunits.Diameter D "Pipe diameter";
    input Modelica.SIunits.Acceleration g "Graviational accelleration";
    input Modelica.SIunits.Distance L "Length of pipe";
    input Real k "loss coefficient";
    input Real f "Friction factor";
    output Real v "Flow velocity";
  algorithm
    v :=  sqrt((H_in-H_out)/((2*g)*(k+f*L/D)));
  end v_Bernoulli;

  function f_Bernoulli
    extends Modelica.Icons.Function;
    input Modelica.SIunits.Height H_in "H. head at reservoir";
    input Modelica.SIunits.Height H_out "H. head at sink";
    input Modelica.SIunits.Diameter D "Pipe diameter";
    input Modelica.SIunits.Acceleration g "Graviational accelleration";
    input Modelica.SIunits.Distance L "Length of pipe";
    input Real v "Friction factor";
    output Real f "Flow velocity";
  algorithm
    f :=  (2*D*g*(H_in-H_out))/(L*V^2);
  end f_Bernoulli;

  function H_f
  extends Modelica.Icons.Function;
    input Modelica.SIunits.Diameter D "Pipe diameter";
    input Modelica.SIunits.Acceleration g "Graviational accelleration";
    input Modelica.SIunits.Distance L "Length of pipe";
    input Modelica.SIunits.Velocity V;
    input Real f "Friction factor";
    output Real H;
  algorithm
    H := f*(L/D)*((V^2)/(2*g));
  end H_f;

  function v_iterate
    extends Modelica.Icons.Function;
    input Modelica.SIunits.Diameter D "Pipe diameter";
    input Modelica.SIunits.Density rho "Density";
    input Modelica.SIunits.DynamicViscosity mu "Dynamic viscosity of water";
    input Modelica.SIunits.Height eps "Pipe roughness height";
    input Modelica.SIunits.Length L "Length";
    input Modelica.SIunits.Acceleration g "Graviational accelleration";
    input Real v_guess "Guessed v using frictionless Bernoulli";
    input Real hf_guess "Change in h due to guessed friction";
    input Real dh "Real dh";
    output Real v_out "Flow velocity";
  protected
    Real f_guess;
  algorithm
    while hf_guess>=dh loop
      v_guess :=v_guess*0.95;
      f_guess :=100*OpenWPL.Functions.f_Haaland(v_guess,D,rho,mu,eps);
      hf_guess:=OpenWPL.Functions.H_f(f_guess,L,D,v_guess,g);
    end while;
    v_out := v_guess;
  end v_iterate;
  
end Functions;
