within OpenWPL;

package Components "Some components that are missing from other libraries"
  extends Modelica.Icons.Package;

  model Pipe_OpenHPL "Dynamic pipe model with pressure surges"
    outer OpenWPL.Data data "Using standard data set";
    extends OpenWPL.Icons.Waterhammer;
    import OpenWPL.Data;
    import Modelica.SIunits;
    import Modelica.Constants;
    // Constants
    import Modelica.Constants.pi;
    //import Data.g;
    // Geometrical parameters of the pipe
    parameter SIunits.Length H = 0 "Height difference from the inlet to the outlet" annotation(
      Dialog(group = "Geometry"));
    parameter SIunits.Length L = 4 "Length of the pipe" annotation(
      Dialog(group = "Geometry"));
    parameter SIunits.Diameter D_i = 4 "Diameter of the inlet side" annotation(
      Dialog(group = "Geometry"));
    parameter SIunits.Diameter D_o = D_i "Diameter of the outlet side" annotation(
      Dialog(group = "Geometry"));
    parameter SIunits.Height eps = 5e-2 "Pipe roughness height" annotation(
      Dialog(group = "Geometry"));
    // Fluid properties in the system
    parameter SIunits.Density rho = 0.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.DynamicViscosity mu = 2e-3 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Initial parameters
    parameter Modelica.SIunits.VolumeFlowRate Vdot_0 = data.V_0 "Initial flow rate" annotation(
      Dialog(group = "Initialization"));
    // Variables
    Modelica.SIunits.Diameter D_ = 0.5 * (D_i + D_o) "Average diameter";
    Modelica.SIunits.Area A_i = D_i ^ 2 * pi / 4 "Inlet cross section area";
    Modelica.SIunits.Area A_o = D_o ^ 2 * pi / 4 "Outlet cross section area";
    Modelica.SIunits.Area A_ = D_ ^ 2 * pi / 4 "Average cross section area";
    Modelica.SIunits.Length dx;
    Modelica.SIunits.Time dt;
    Modelica.SIunits.Velocity v "Water velocity";
    Real f;
    Real B;
    Real R;
    Modelica.SIunits.VolumeFlowRate Vdot(start = Vdot_0) "Flow rate";
    extends OpenHPL.Interfaces.ContactPort;
  equation
    Vdot = mdot / rho;
    v = Vdot / A_;
    dx = L / 3;
    dt = dx / a;
    f = Functions.Haaland(v, D_, rho, mu, eps);
    B = Functions.Impedance(a, data.g, A_);
//B = a/(g*A_);
    R = Functions.Resistance(f, dx, Data.g, D_, A_);
//R = (f*dx)/(2*g*D_*A_^2);
    annotation(
      defaultComponentName = "pipe",
      Icon(coordinateSystem(preserveAspectRatio = false, initialScale = 0.1)),
      Documentation(info = "<html>
  <p>
  A pipe using the method of characteristics in order to simulate pressure surges.
  </p>
  </html>"),
      Diagram(coordinateSystem(initialScale = 0.1)));
  end Pipe_OpenHPL;

  model Pipe_MSL_v1
    /*
            model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.

            author:   Jonas Tveit Hinna
            date:     03.03.2021
            version:  1
            */
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    //Medium.BaseProperties medium;
    String name = getInstanceName();
    // Geometry
    parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter "Diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area crossArea = Modelica.Constants.pi * diameter * diameter / 4 "Inner cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Length perimeter(min = 0) = Modelica.Constants.pi * diameter "Inner perimeter" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = crossArea * length * nParallel "volume size";
    // Static head
    parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Method of Characteristics declarations
    //Real f "Darcy-Weisbach friction factor";
    //Real B "Impedance from MOC";
    //Real R "Restistance from MOC";
    // Local vector storage. Global values are stored in the component "MOC.mo"
    //Real f_vector[3];
    //Real B_vector[3];
    //Real R_vector[3];
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    //constant Real rho = 0.2;
    final parameter SIunits.Density rho = 0.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 2e-3 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Variables
    //Real dx "Length step calculated in this model";
    //Real dx_min = moc.dx_min;
    //Real dt "Time step";
    //Real v "Water velocity";
    final parameter Integer N = 4 "n parts";
  equation
/*
  dx = length/3;
  //moc.time_step[1] = dx; not valid anymore
  dt = dx/a;
  //N = length/dx;
  
  for i in 2:N loop
    //f = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
    f = 0; //testing for a square pressure pulse
    B = OpenWPL.Functions.Impedance(a=a, A=crossArea, g=system.g);
    R = OpenWPL.Functions.Resistance(f=f, dx=dx, g=system.g, D=diameter, A=crossArea);
  end for;
  */
//port_a.m_flow = port_b.m_flow;
//0 = port_a.m_flow + port_b.m_flow;
//port_a.h_outflow = port_b.h_outflow;
//port_a.m_flow = flowModel.m_flows[1];
//0 = port_a.m_flow + port_b.m_flow;
    port_a.m_flow = system.m_flow_start;
    port_b.m_flow = system.m_flow_start;
//0 = port_a.m_flow + port_b.m_flow;
    port_a.Xi_outflow = inStream(port_b.Xi_outflow);
    port_b.Xi_outflow = inStream(port_a.Xi_outflow);
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
    port_b.h_outflow = inStream(port_a.h_outflow) - system.g * height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow) + system.g * height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
<p>Model of a straight pipe modelled using the method of characteristics.
</p>
</html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian");
  end Pipe_MSL_v1;

  model Pipe_MSL_v2
    /*
              model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              author:   Jonas Tveit Hinna
              date:     16.03.2021
              version:  2
              */
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    //Medium.BaseProperties medium;
    String name = getInstanceName();
    // Geometry
    parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter "Diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area crossArea = Modelica.Constants.pi * diameter * diameter / 4 "Inner cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Length perimeter(min = 0) = Modelica.Constants.pi * diameter "Inner perimeter" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = crossArea * length * nParallel "volume size";
    // Static head
    parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 0.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 2e-3 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    final parameter Real f "Darcy-Weisbach friction factor";
    final parameter Real B "Impedance from MOC";
    final parameter Real R "Resistance from MOC";
    final parameter Real dx = length / 3 "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    final parameter Real v "Water velocity";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    // Local array storage. Global values are stored in the component "MOC.mo"
    Real fp_array[time, N](each start = 0);
    Real fm_array[time, N](each start = 0);
    Real Rp_array[time, N](each start = 0);
    Real Rm_array[time, N](each start = 0);
    Real Bp_array[time, N](each start = 0);
    Real Bm_array[time, N](each start = 0);
    Real Cp_array[time, N](each start = 0);
    Real Cm_array[time, N](each start = 0);
    // Local array storage.
    Real f_array[time, N](each start = 0);
    Real R_array[time, N](each start = 0);
    Real B_array[time, N](each start = 0);
    Real Cp_array[time, N](each start = 0);
    Real Cm_array[time, N](each start = 0);
    Real H_array[time, N](each start = 0);
    Real Q_array[time, N](each start = 0);
    // Initial conditions
    final parameter SI.VolumeFlowRate Qa = system.m_flow_start / rho;
    final parameter SI.VolumeFlowRate Qb = Qa;
    final parameter SI.Pressure Ha = system.p_start;
    final parameter SI.Pressure Hb = Ha;
  equation
//moc.time_step[1] = dx; not valid anymore
// Initial conditions for the plus characteristic
    fp_array[1, :] = 0;
    Rp_array[1, :] = OpenWPL.Functions.Resistance(f = f, dx = dx_min, g = system.g, D = diameter, A = crossArea);
    Bp_array[1, :] = OpenWPL.Functions.Impedance(a = a, A = crossArea, g = system.g);
    Cp_array[1, :] = OpenWPL.Functions.Cplus(Hb = Hb, B = Bp_array[1, 1], Qb = Qb, R = Rp_array[1, 1]);
// Initial conditions for the minus characteristic
    fm_array[1, :] = 0;
    Rm_array[1, :] = OpenWPL.Functions.Resistance(f = f, dx = dx_min, g = system.g, D = diameter, A = crossArea);
    Bm_array[1, :] = Bp_array;
    Cm_array[1, :] = OpenWPL.Functions.Cminus(Ha = Ha, B = Bm_array[1, 1], Qa = Qa, R = Rm_array[1, 1]);
// Initial conditions
    f_array[1, :] = 0;
    R_array[1, :] = OpenWPL.Functions.Resistance(f = f, dx = dx_min, g = system.g, D = diameter, A = crossArea);
    Bm_array[1, :] = Bp_array;
    Cm_array[1, :] = OpenWPL.Functions.Cminus(Ha = Ha, B = Bm_array[1, 1], Qa = Qa, R = Rm_array[1, 1]);
// Initial H and Q
    H_array[1, :] = OpenWPL.Functions.Hpoint(Cp = Cp_array[1, 1], Cm = Cm_array[1, 3]);
//Q_array[1,:];
// Time iteration loop
    for j in 1:time loop
// BC V
// BC H
// Space interation loop
      for i in 2:N - 1 loop
//f = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
        fp_array[j, i] = 0;
//testing for a square pressure pulse
        Bp_array[j, i] = OpenWPL.Functions.Impedance(a = a, A = crossArea, g = system.g);
        Rp_array[j, i] = OpenWPL.Functions.Resistance(f = f, dx = dx_min, g = system.g, D = diameter, A = crossArea);
//Cp_vector[j,i] = OpenWPL.Functions.Cplus();
//Cm_vector[j,i] = OpenWPL.Functions.Cminuns();
      end for;
    end for;
// Port declarations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    port_b.m_flow = system.m_flow_start;
// Port declarations for substance mass flow
    port_a.Xi_outflow = inStream(port_b.Xi_outflow);
    port_b.Xi_outflow = inStream(port_a.Xi_outflow);
// Port declarations for trace substance mass flow
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
// Port declarations for enthalpy flow
    port_b.h_outflow = inStream(port_a.h_outflow) - system.g * height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow) + system.g * height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian");
  end Pipe_MSL_v2;

  model Pipe_MSL_v3
    /*
              model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              author:   Jonas Tveit Hinna
              date:     16.03.2021
              version:  3
              */
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    //Medium.BaseProperties medium;
    //String name = getInstanceName();
    // Geometry
    parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter "Diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area crossArea = Modelica.Constants.pi * diameter * diameter / 4 "Inner cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Length perimeter(min = 0) = Modelica.Constants.pi * diameter "Inner perimeter" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = crossArea * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 0.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 2e-3 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    Real f "Darcy-Weisbach friction factor";
    Real B "Impedance from MOC";
    Real R "Resistance from MOC";
    final parameter Real dx = length / 3 "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    Real v "Water velocity";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    // Temporary time constant
    final parameter Integer T = 2;
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    final parameter Real fp_array[T, N];
    final parameter Real fm_array[T, N];
    final parameter Real Rp_array[T, N];
    final parameter Real Rm_array[T, N];
    final parameter Real Bp_array[T, N];
    final parameter Real Bm_array[T, N];
    final parameter Real Cp_array[T, N];
    final parameter Real Cm_array[T, N];
    final parameter Real H_array[T, N];
    final parameter Real Q_array[T, N];
    // Initial condition declarations
    final parameter SI.VolumeFlowRate Qa = system.m_flow_start / rho;
    final parameter SI.VolumeFlowRate Qb = Qa;
    final parameter SI.Length Ha = system.p_start / (rho * system.g);
    final parameter SI.Length Hb = Ha;
  initial equation
//moc.time_step[1] = dx; not imlemented yet
// Initial conditions for the plus characteristic
    fp_array[1, :] = {0.0 * k for k in 1:N};
    Rp_array[1, :] = {OpenWPL.Functions.R_Resistance(f = fp_array[1, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea) * k for k in 1:N};
    Bp_array[1, :] = {OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) * k for k in 1:N};
    Cp_array[1, :] = {OpenWPL.Functions.C_plus(Ha = Ha, B = Bp_array[1, 1], Qa = Qa) * k for k in 1:N};
// Initial conditions for the minus characteristic
    fm_array[1, :] = {0.0 * k for k in 1:N};
    Rm_array[1, :] = {OpenWPL.Functions.R_Resistance(f = fm_array[1, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea) * k for k in 1:N};
    Bm_array[1, :] = {OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) * k for k in 1:N};
    Cm_array[1, :] = {OpenWPL.Functions.C_minus(Hb = Hb, B = Bm_array[1, 1], Qb = Qb) * k for k in 1:N};
// Initial conditions for H and Q
    H_array[1, :] = {OpenWPL.Functions.H_head(Cp = Cp_array[1, 1], Cm = Cm_array[1, 1], Bm = Bm_array[1, 1], Bp = Bp_array[1, 1]) * k for k in 1:N};
    Q_array[1, :] = {OpenWPL.Functions.Q_flow(Cp = Cp_array[1, 1], Cm = Cm_array[1, 1], Bm = Bm_array[1, 1], Bp = Bp_array[1, 1]) * k for k in 1:N};
  equation
// Time iteration loop
    for j in 1:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
      Cm_array[j + 1, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 2], B = Bm_array[j, 1], Qb = Q_array[j, 2]);
// H_array[j+1,1] = Source.ports[1].p/(rho*system.g);
      H_array[j + 1, 1] = 1.0;
      Q_array[j + 1, 1] = (Cm_array[j + 1, 1] - H_array[j + 1, 1]) / Bm_array[j, 1];
// BC right: closed by valve with Q=0
      Cp_array[j + 1, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N - 1], B = Bp_array[j, N], Qa = Q_array[j, N - 1]);
      Q_array[j + 1, N] = 0.0;
      H_array[j + 1, N] = Cp_array[j + 1, N];
// Space iteration loop
      for i in 2:N - 1 loop
//fp = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
        fp_array[j, i] = 0.0;
//testing for a square pressure pulse
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, i] * abs(Q_array[j, i - 1]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i - 1], B = Bp_array[j, i], Qa = Q_array[j, i - 1]);
//fm = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
        fm_array[j, i] = 0.0;
//testing for a square pressure pulse
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, i] * abs(Q_array[j, i + 1]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i + 1], B = Bm_array[j, i], Qb = Q_array[j, i + 1]);
        H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j, i], Cm = Cm_array[j, i], Bm = Bm_array[j, i], Bp = Bp_array[j, i])});
        Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j, i], Cm = Cm_array[j, i], Bm = Bm_array[j, i], Bp = Bp_array[j, i])});
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    port_b.m_flow = system.m_flow_start;
// Port equations for substance mass flow
    port_a.Xi_outflow = inStream(port_b.Xi_outflow);
    port_b.Xi_outflow = inStream(port_a.Xi_outflow);
// Port equations for trace substance mass flow
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
// Port equations for enthalpy flow
    port_b.h_outflow = inStream(port_a.h_outflow) - system.g * height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow) + system.g * height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v3;

  model Pipe_MSL_v4
    /*
              model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              author:   Jonas Tveit Hinna
              date:     16.03.2021
              version:  4
              */
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    //Medium.BaseProperties medium;
    //String name = getInstanceName();
    // Geometry
    parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter "Diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area crossArea = Modelica.Constants.pi * diameter * diameter / 4 "Inner cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Length perimeter(min = 0) = Modelica.Constants.pi * diameter "Inner perimeter" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = crossArea * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 0.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 2e-3 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    Real f "Darcy-Weisbach friction factor";
    Real B "Impedance from MOC";
    Real R "Resistance from MOC";
    final parameter Real dx = length / 6 "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    Real v "Water velocity";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    // Temporary time constant
    parameter Integer T = 6;
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real fp_array[T + 1, N];
    Real fm_array[T + 1, N];
    Real Rp_array[T + 1, N];
    Real Rm_array[T + 1, N];
    Real Bp_array[T + 1, N];
    Real Bm_array[T + 1, N];
    Real Cp_array[T + 1, N];
    Real Cm_array[T + 1, N];
    Real H_array[T + 1, N];
    Real Q_array[T + 1, N];
    // Initial condition declarations
    final parameter SI.VolumeFlowRate Qa = system.m_flow_start / rho;
    final parameter SI.VolumeFlowRate Qb = Qa;
    final parameter SI.Length Ha = system.p_start / (rho * system.g);
    final parameter SI.Length Hb = Ha;
  equation
//moc.time_step[1] = dx; not imlemented yet
// Initial conditions for the plus characteristic
    fp_array[1, :] = {0.0 * k for k in 1:N};
    Rp_array[1, :] = {OpenWPL.Functions.R_Resistance(f = fp_array[1, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea) * k for k in 1:N};
    Bp_array[1, :] = {OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) * k for k in 1:N};
    Cp_array[1, :] = {OpenWPL.Functions.C_plus(Ha = Ha, B = Bp_array[1, 1], Qa = Qa) * k for k in 1:N};
// Initial conditions for the minus characteristic
    fm_array[1, :] = {0.0 * k for k in 1:N};
    Rm_array[1, :] = {OpenWPL.Functions.R_Resistance(f = fm_array[1, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea) * k for k in 1:N};
    Bm_array[1, :] = {OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) * k for k in 1:N};
    Cm_array[1, :] = {OpenWPL.Functions.C_minus(Hb = Hb, B = Bm_array[1, 1], Qb = Qb) * k for k in 1:N};
// Initial conditions for H and Q
    H_array[1, :] = {OpenWPL.Functions.H_head(Cp = Cp_array[1, 1], Cm = Cm_array[1, 1], Bm = Bm_array[1, 1], Bp = Bp_array[1, 1]) * k for k in 1:N};
    Q_array[1, :] = {OpenWPL.Functions.Q_flow(Cp = Cp_array[1, 1], Cm = Cm_array[1, 1], Bm = Bm_array[1, 1], Bp = Bp_array[1, 1]) * k for k in 1:N};
// Time iteration loop
    for j in 1:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
      Cm_array[j + 1, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 2], B = Bm_array[j, 1], Qb = Q_array[j, 2]);
// H_array[j+1,1] = Source.ports[1].p/(rho*system.g);
      H_array[j + 1, 1] = 1.0;
      Q_array[j + 1, 1] = (H_array[j + 1, 1] - Cm_array[j + 1, 1]) / Bm_array[j, 1];
// BC right: closed by valve with Q=0
      Cp_array[j + 1, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N - 1], B = Bp_array[j, N], Qa = Q_array[j, N - 1]);
      Q_array[j + 1, N] = 0.0;
      H_array[j + 1, N] = Cp_array[j + 1, N] - Bp_array[j, N] * Q_array[j + 1, N];
// Space iteration loop
      for i in 2:N - 1 loop
//fp = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
        fp_array[j + 1, i] = 0.0;
//testing for a square pressure pulse
        Rp_array[j + 1, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bp_array[j + 1, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, i] * abs(Q_array[j, i - 1]);
        Cp_array[j + 1, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i - 1], B = Bp_array[j, i], Qa = Q_array[j, i - 1]);
//fm = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
        fm_array[j + 1, i] = 0.0;
//testing for a square pressure pulse
        Rm_array[j + 1, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bm_array[j + 1, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, i] * abs(Q_array[j, i + 1]);
        Cm_array[j + 1, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i + 1], B = Bm_array[j, i], Qb = Q_array[j, i + 1]);
        H_array[j + 1, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j, i], Cm = Cm_array[j, i], Bm = Bm_array[j, i], Bp = Bp_array[j, i])});
        Q_array[j + 1, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j, i], Cm = Cm_array[j, i], Bm = Bm_array[j, i], Bp = Bp_array[j, i])});
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    port_b.m_flow = system.m_flow_start;
// Port equations for substance mass flow
    port_a.Xi_outflow = inStream(port_b.Xi_outflow);
    port_b.Xi_outflow = inStream(port_a.Xi_outflow);
// Port equations for trace substance mass flow
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
// Port equations for enthalpy flow
    port_b.h_outflow = inStream(port_a.h_outflow) - system.g * height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow) + system.g * height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v4;

  model Pipe_MSL_v5
    /*
              model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              author:   Jonas Tveit Hinna
              date:     30.03.2021
              version:  5
              */
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    //Medium.BaseProperties medium;
    //String name = getInstanceName();
    // Geometry
    parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter "Diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area crossArea = Modelica.Constants.pi * diameter * diameter / 4 "Inner cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Length perimeter(min = 0) = Modelica.Constants.pi * diameter "Inner perimeter" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = crossArea * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 0.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 2e-3 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    //Real f "Darcy-Weisbach friction factor";
    //Real B "Impedance from MOC";
    //Real R "Resistance from MOC";
    final parameter Real dx = length / 5 "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //Real v "Water velocity";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    // Temporary time constant
    parameter Integer T = 4;
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real fp_array[T, N];
    Real fm_array[T, N];
    Real Rp_array[T, N];
    Real Rm_array[T, N];
    Real Bp_array[T, N];
    Real Bm_array[T, N];
    Real Cp_array[T, N];
    Real Cm_array[T, N];
    Real H_array[T, N];
    Real Q_array[T, N];
    // Initial condition declarations
    //final parameter SI.Velocity v = sqrt(2*system.g*Source.level);
    //final parameter SI.VolumeFlowRate Qa = v*crossArea;
    //final parameter SI.VolumeFlowRate Qb = Qa;
    //final parameter SI.Length Ha = Source.level;
    //final parameter SI.Length Hb = Ha - 1;
    final parameter SI.VolumeFlowRate Qa = 1;
    final parameter SI.VolumeFlowRate Qb = Qa;
    final parameter SI.Length Ha = 10;
    final parameter SI.Length Hb = Ha;
  equation
//moc.time_step[1] = dx; not imlemented yet
// Initial condition loop
    for k in 1:N loop
      H_array[1, k] = Ha;
      Q_array[1, k] = sqrt(2 * system.g * Ha) * crossArea;
// Initial conditions for the plus characteristic
      fp_array[1, k] = 0.0;
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = Ha, B = Bp_array[1, 1], Qa = Q_array[1, k]);
// Initial conditions for the minus characteristic
      fm_array[1, k] = 0.0;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = Hb, B = Bm_array[1, 1], Qb = Q_array[1, k]);
// Initial conditions for H and Q
//H_array[1, k] = OpenWPL.Functions.H_head(Cp = Cp_array[1, 1], Cm = Cm_array[1, 1], Bm = Bm_array[1, 1], Bp = Bp_array[1, 1]);
//Q_array[1, k] = OpenWPL.Functions.Q_flow(Cp = Cp_array[1, 1], Cm = Cm_array[1, 1], Bm = Bm_array[1, 1], Bp = Bp_array[1, 1]);
    end for;
// End I.C loop
// Time iteration loop
    for j in 2:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
//Cm_array[j,1] = OpenWPL.Functions.C_minus(Hb=H_array[j-1,2], B=Bm_array[j-1,1], Qb=Q_array[j-1,2]);
// H_array[j+1,1] = Source.ports[1].p/(rho*system.g);
      H_array[j, 1] = 10.0;
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 1]) / Bm_array[j - 1, 1];
//fp_array[j, i] = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
      fp_array[j, 1] = 0.0;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = Bp_array[j, 1], Qa = Q_array[j, 1]);
//fm = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
      fm_array[j, 1] = 0.0;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = Bm_array[j, 1], Qb = Q_array[j, 1]);
// BC right: closed by valve with Q=0
//Cp_array[j,N] = OpenWPL.Functions.C_plus(Ha=H_array[j,N-1], B=Bp_array[j,N], Qa=Q_array[j,N-1]);
      Q_array[j, N] = 0.0;
      H_array[j, N] = Cp_array[j - 1, N] - Bp_array[j - 1, N] * Q_array[j, N];
//fp_array[j, i] = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
      fp_array[j, N] = 0.0;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = Bp_array[j, N], Qa = Q_array[j, N]);
//fm = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
      fm_array[j, N] = 0.0;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = Bm_array[j, N], Qb = Q_array[j, N]);
// Space iteration loop
      for i in 2:N - 1 loop
        H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
//fp_array[j, i] = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
        fp_array[j, i] = 0.0;
//testing for a square pressure pulse
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, i] * abs(Q_array[j, i]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = Bp_array[j, i], Qa = Q_array[j, i]);
//fm = OpenWPL.Functions.Haaland(v=v, D=diameter, rho=rho, mu=mu, eps=roughness);
        fm_array[j, i] = 0.0;
//testing for a square pressure pulse
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, i] * abs(Q_array[j, i]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = Bm_array[j, i], Qb = Q_array[j, i]);
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    port_b.m_flow = system.m_flow_start;
// Port equations for substance mass flow
    port_a.Xi_outflow = inStream(port_b.Xi_outflow);
    port_b.Xi_outflow = inStream(port_a.Xi_outflow);
// Port equations for trace substance mass flow
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
// Port equations for enthalpy flow
    port_b.h_outflow = inStream(port_a.h_outflow) - system.g * height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow) + system.g * height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v5;

  model Pipe_MSL_v6
    /*
              model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              author:   Jonas Tveit Hinna
              date:     19.04.2021
              version:  6
              */
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    //Medium.BaseProperties medium;
    //String name = getInstanceName();
    // Geometry
    parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter "Diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area crossArea = Modelica.Constants.pi * diameter * diameter / 4 "Inner cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Length perimeter(min = 0) = Modelica.Constants.pi * diameter "Inner perimeter" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = crossArea * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 0.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 2e-3 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    //Real f "Darcy-Weisbach friction factor";
    //Real B "Impedance from MOC";
    //Real R "Resistance from MOC";
    final parameter Real dx = length / 5 "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //Real v "Water velocity";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    // Temporary time constant
    parameter Integer T = 15;
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real fp_array[T, N];
    Real fm_array[T, N];
    Real Rp_array[T, N];
    Real Rm_array[T, N];
    Real Bp_array[T, N];
    Real Bm_array[T, N];
    Real Cp_array[T, N];
    Real Cm_array[T, N];
    Real H_array[T, N];
    Real Q_array[T, N];
    // Initial condition declarations
    final parameter SI.Velocity v = sqrt(2 * system.g * 10);
    //final parameter SI.VolumeFlowRate Qa = v*crossArea;
    //final parameter SI.VolumeFlowRate Qb = Qa;
    //final parameter SI.Length Ha = Source.level;
    //final parameter SI.Length Hb = Ha - 1;
    final parameter SI.VolumeFlowRate Qa = 1;
    final parameter SI.VolumeFlowRate Qb = Qa;
    final parameter SI.Length Ha = 10;
    final parameter SI.Length Hb = Ha;
  equation
//moc.time_step[1] = dx; not imlemented yet
// Initial condition loop
    for k in 1:N loop
      H_array[1, k] = Ha;
      Q_array[1, k] = sqrt(2 * system.g * Ha) * crossArea;
// Initial conditions for the plus characteristic
      fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v, D = diameter, rho = rho, mu = mu, eps = roughness);
//fp_array[1, k] = 0.0;
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = Ha, B = Bp_array[1, 1], Qa = Q_array[1, k]);
// Initial conditions for the minus characteristic
      fm_array[1, k] = OpenWPL.Functions.f_Haaland(v = v, D = diameter, rho = rho, mu = mu, eps = roughness);
//fm_array[1, k] = 0.0;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = Hb, B = Bm_array[1, 1], Qb = Q_array[1, k]);
// Initial conditions for H and Q
//H_array[1, k] = OpenWPL.Functions.H_head(Cp = Cp_array[1, 1], Cm = Cm_array[1, 1], Bm = Bm_array[1, 1], Bp = Bp_array[1, 1]);
//Q_array[1, k] = OpenWPL.Functions.Q_flow(Cp = Cp_array[1, 1], Cm = Cm_array[1, 1], Bm = Bm_array[1, 1], Bp = Bp_array[1, 1]);
    end for;
// End I.C loop
// Time iteration loop
    for j in 2:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
//Cm_array[j,1] = OpenWPL.Functions.C_minus(Hb=H_array[j-1,2], B=Bm_array[j-1,1], Qb=Q_array[j-1,2]);
// H_array[j+1,1] = Source.ports[1].p/(rho*system.g);
      H_array[j, 1] = 10.0;
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 2]) / Bm_array[j - 1, 2];
      fp_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v, D = diameter, rho = rho, mu = mu, eps = roughness);
//fp_array[j,1] = 0.0;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = Bp_array[j, 1], Qa = Q_array[j, 1]);
      fm_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v, D = diameter, rho = rho, mu = mu, eps = roughness);
//fm_array[j,1] = 0.0;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = Bm_array[j, 1], Qb = Q_array[j, 1]);
// BC right: closed by valve with Q=0
//Cp_array[j,N] = OpenWPL.Functions.C_plus(Ha=H_array[j,N-1], B=Bp_array[j,N], Qa=Q_array[j,N-1]);
      Q_array[j, N] = 0.0;
      H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
      fp_array[j, N] = OpenWPL.Functions.f_Haaland(v = v, D = diameter, rho = rho, mu = mu, eps = roughness);
//fp_array[j,N] = 0.0;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = Bp_array[j, N], Qa = Q_array[j, N]);
      fm_array[j, N] = OpenWPL.Functions.f_Haaland(v = v, D = diameter, rho = rho, mu = mu, eps = roughness);
//fm_array[j,N] = 0.0;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = Bm_array[j, N], Qb = Q_array[j, N]);
// Space iteration loop
      for i in 2:N - 1 loop
        H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        fp_array[j, i] = OpenWPL.Functions.f_Haaland(v = v, D = diameter, rho = rho, mu = mu, eps = roughness);
//fp_array[j,i] = 0.0; //testing for a square pressure pulse
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, i] * abs(Q_array[j, i]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = Bp_array[j, i], Qa = Q_array[j, i]);
        fm_array[j, i] = OpenWPL.Functions.f_Haaland(v = v, D = diameter, rho = rho, mu = mu, eps = roughness);
//fm_array[j,i] = 0.0; //testing for a square pressure pulse
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, i] * abs(Q_array[j, i]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = Bm_array[j, i], Qb = Q_array[j, i]);
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    port_b.m_flow = system.m_flow_start;
// Port equations for substance mass flow
    port_a.Xi_outflow = inStream(port_b.Xi_outflow);
    port_b.Xi_outflow = inStream(port_a.Xi_outflow);
// Port equations for trace substance mass flow
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
// Port equations for enthalpy flow
    port_b.h_outflow = inStream(port_a.h_outflow) - system.g * height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow) + system.g * height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v6;

  model Pipe_MSL_v7
    /*
              model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              author:   Jonas Tveit Hinna
              date:     19.04.2021
              version:  7
            */
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    //Medium.BaseProperties medium;
    //String name = getInstanceName();
    // Geometry
    parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter "Diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area crossArea = Modelica.Constants.pi * diameter * diameter / 4 "Inner cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Length perimeter(min = 0) = Modelica.Constants.pi * diameter "Inner perimeter" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = crossArea * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 0.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 2e-3 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    parameter Boolean friction = true "= true if there is friction in the system" annotation(
      Dialog(enable = true, tab = "General"));
    final parameter Real dx = length / moc.N "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //Real v "Water velocity";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    // Temporary time constant
    parameter Integer T = moc.T;
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real v_array[T, N];
    Real fp_array[T, N];
    Real fm_array[T, N];
    Real Rp_array[T, N];
    Real Rm_array[T, N];
    Real Bp_array[T, N];
    Real Bm_array[T, N];
    Real Cp_array[T, N];
    Real Cm_array[T, N];
    Real H_array[T, N];
    Real Q_array[T, N];
    // Initial condition declarations
    final parameter SI.VolumeFlowRate Qa = 1;
    final parameter SI.VolumeFlowRate Qb = Qa;
    final parameter SI.Length Ha = 10;
    final parameter SI.Length Hb = Ha;
    final parameter SI.Velocity v = sqrt(2 * system.g * Ha);
    //final parameter SI.VolumeFlowRate Qa = v*crossArea;
    //final parameter SI.VolumeFlowRate Qb = Qa;
    //final parameter SI.Length Ha = Source.level;
    //final parameter SI.Length Hb = Ha - 1;
  equation
//moc.time_step[1] = dx; not imlemented yet
// Initial condition loop
    for k in 1:N loop
      H_array[1, k] = Ha;
      v_array[1, k] = OpenWPL.Functions.v_Velocity(H = H_array[1, k], g = system.g);
      Q_array[1, k] = v_array[1, k] * crossArea;
// Initial conditions for the plus characteristic
      if friction then
        fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[1, k] = 0.0;
      end if;
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, k], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = H_array[1, k], B = Bp_array[1, k], Qa = Q_array[1, k]);
// Initial conditions for the minus characteristic
      if friction then
        fm_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[1, k] = 0.0;
      end if;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, k], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = Hb, B = Bm_array[1, k], Qb = Q_array[1, k]);
    end for;
// End I.C loop
// Time iteration loop
    for j in 2:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
//Cm_array[j,1] = OpenWPL.Functions.C_minus(Hb=H_array[j-1,2], B=Bm_array[j-1,1], Qb=Q_array[j-1,2]);
// H_array[j+1,1] = Source.ports[1].p/(rho*system.g);
      H_array[j, 1] = Ha;
      v_array[j, 1] = OpenWPL.Functions.v_Velocity(H = H_array[j, 1], g = system.g);
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 2]) / Bm_array[j - 1, 2];
      if friction then
        fp_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, 1] = 0.0;
      end if;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = Bp_array[j, 1], Qa = Q_array[j, 1]);
      if friction then
        fm_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, 1] = 0.0;
      end if;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = Bm_array[j, 1], Qb = Q_array[j, 1]);
// BC right: closed by valve with Q=0
//Cp_array[j,N] = OpenWPL.Functions.C_plus(Ha=H_array[j,N-1], B=Bp_array[j,N], Qa=Q_array[j,N-1]);
      if closedValve then
        Q_array[j, N] = 0.0;
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
      end if;
      if friction then
        fp_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, N] = 0.0;
      end if;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = Bp_array[j, N], Qa = Q_array[j, N]);
      if friction then
        fm_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, N] = 0.0;
      end if;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = Bm_array[j, N], Qb = Q_array[j, N]);
// Space iteration loop
      for i in 2:N - 1 loop
        H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        v_array[j, i] = OpenWPL.Functions.v_Velocity(H = H_array[j, i], g = system.g);
        Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        if friction then
          fp_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = diameter, rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fp_array[j, i] = 0.0;
//testing for a square pressure pulse
        end if;
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, i] * abs(Q_array[j, i]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = Bp_array[j, i], Qa = Q_array[j, i]);
        if friction then
          fm_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = diameter, rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fm_array[j, i] = 0.0;
//testing for a square pressure pulse
        end if;
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, i] * abs(Q_array[j, i]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = Bm_array[j, i], Qb = Q_array[j, i]);
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
//port_b.m_flow = system.m_flow_start;
    0 = port_a.m_flow + port_b.m_flow;
// Port equations for substance mass flow
    port_a.Xi_outflow = inStream(port_b.Xi_outflow);
    port_b.Xi_outflow = inStream(port_a.Xi_outflow);
// Port equations for trace substance mass flow
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
// Port equations for enthalpy flow
    port_b.h_outflow = inStream(port_a.h_outflow) - system.g * height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow) + system.g * height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v7;

  model Pipe_MSL_v8
    /*
              model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              author:   Jonas Tveit Hinna
              date:     19.04.2021
              version:  8
            */
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    //Medium.BaseProperties medium;
    //String name = getInstanceName();
    // Geometry
    final parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    final parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter_in "Inlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    parameter SI.Diameter diameter_out "Outlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area crossArea = Modelica.Constants.pi * diameter * diameter / 4 "Inner cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Length perimeter(min = 0) = Modelica.Constants.pi * diameter "Inner perimeter" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = crossArea * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    final parameter SI.Diameter diameter = diameter_in;
    // Static head
    parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 997 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 8.9e-5 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    parameter Boolean friction = true "= true if there is friction in the system" annotation(
      Dialog(enable = true, tab = "General"));
    final parameter Real dx = length / moc.N "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //Real v "Water velocity";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    // Temporary time constant
    parameter Integer T = moc.T;
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real v_array[T, N];
    Real fp_array[T, N];
    Real fm_array[T, N];
    Real Rp_array[T, N];
    Real Rm_array[T, N];
    Real Bp_array[T, N];
    Real Bm_array[T, N];
    Real Cp_array[T, N];
    Real Cm_array[T, N];
    Real H_array[T, N];
    Real Q_array[T, N];
    // Initial condition declarations
    final parameter SI.VolumeFlowRate Qa = 1;
    final parameter SI.VolumeFlowRate Qb = Qa;
    final parameter SI.Length Ha = 10;
    final parameter SI.Length Hb = Ha;
    final parameter SI.Velocity v = sqrt(2 * system.g * Ha);
    //final parameter SI.VolumeFlowRate Qa = v*crossArea;
    //final parameter SI.VolumeFlowRate Qb = Qa;
    //final parameter SI.Length Ha = Source.level;
    //final parameter SI.Length Hb = Ha - 1;
  equation
//moc.time_step[1] = dx; not imlemented yet
// Initial condition loop
    for k in 1:N loop
      H_array[1, k] = Ha;
      v_array[1, k] = OpenWPL.Functions.v_Velocity(H = H_array[1, k], g = system.g);
      Q_array[1, k] = v_array[1, k] * crossArea;
// Initial conditions for the plus characteristic
      if friction then
        fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[1, k] = 0.0;
      end if;
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, k], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[1, k] * abs(Q_array[1, k]);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = H_array[1, k], B = Bp_array[1, k], Qa = Q_array[1, k]);
// Initial conditions for the minus characteristic
      if friction then
        fm_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[1, k] = 0.0;
      end if;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, k], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[1, k] * abs(Q_array[1, k]);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = Hb, B = Bm_array[1, k], Qb = Q_array[1, k]);
    end for;
// End I.C loop
// Time iteration loop
    for j in 2:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
//Cm_array[j,1] = OpenWPL.Functions.C_minus(Hb=H_array[j-1,2], B=Bm_array[j-1,1], Qb=Q_array[j-1,2]);
// H_array[j+1,1] = Source.ports[1].p/(rho*system.g);
      H_array[j, 1] = Ha;
      v_array[j, 1] = OpenWPL.Functions.v_Velocity(H = H_array[j, 1], g = system.g);
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 2]) / Bm_array[j - 1, 2];
      if friction then
        fp_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, 1] = 0.0;
      end if;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = Bp_array[j, 1], Qa = Q_array[j, 1]);
      if friction then
        fm_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, 1] = 0.0;
      end if;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = Bm_array[j, 1], Qb = Q_array[j, 1]);
// BC right: closed by valve with Q=0
//Cp_array[j,N] = OpenWPL.Functions.C_plus(Ha=H_array[j,N-1], B=Bp_array[j,N], Qa=Q_array[j,N-1]);
      if closedValve then
        Q_array[j, N] = 0.0;
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
      end if;
      if friction then
        fp_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, N] = 0.0;
      end if;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = Bp_array[j, N], Qa = Q_array[j, N]);
      if friction then
        fm_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = diameter, rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, N] = 0.0;
      end if;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = diameter, A = crossArea);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = Bm_array[j, N], Qb = Q_array[j, N]);
// Space iteration loop
      for i in 2:N - 1 loop
        H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        v_array[j, i] = scalar(OpenWPL.Functions.v_Velocity(H = H_array[j, i], g = system.g));
        Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        if friction then
          fp_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = diameter, rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fp_array[j, i] = 0.0;
//testing for a square pressure pulse
        end if;
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rp_array[j, i] * abs(Q_array[j - 1, i - 1]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = Bp_array[j, i], Qa = Q_array[j, i]);
        if friction then
          fm_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = diameter, rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fm_array[j, i] = 0.0;
//testing for a square pressure pulse
        end if;
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = diameter, A = crossArea);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = crossArea, g = system.g) + Rm_array[j, i] * abs(Q_array[j - 1, i + 1]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = Bm_array[j, i], Qb = Q_array[j, i]);
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
//port_b.m_flow = system.m_flow_start;
    0 = port_a.m_flow + port_b.m_flow;
// Port equations for substance mass flow
    port_a.Xi_outflow = inStream(port_b.Xi_outflow);
    port_b.Xi_outflow = inStream(port_a.Xi_outflow);
// Port equations for trace substance mass flow
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
// Port equations for enthalpy flow
    port_b.h_outflow = inStream(port_a.h_outflow) - system.g * height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow) + system.g * height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v8;

  model Pipe_MSL_v9
    /*
              Model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              Author:   Jonas Tveit Hinna
              Date:     23.04.2021
              Version:  9
            */
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    // Geometry
    final parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    final parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter_in "Inlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    //final parameter SI.Diameter diameter = diameter_in;
    parameter SI.Diameter diameter_out = diameter_in "Outlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    //final parameter SI.Area crossArea = Modelica.Constants.pi * diameter * diameter / 4 "Inner cross section area" annotation(
    //  Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Area A_in = Modelica.Constants.pi * diameter_in * diameter_in / 4 "Inlet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Area A_out = Modelica.Constants.pi * diameter_out * diameter_out / 4 "Oulet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    //final parameter SI.Length perimeter(min = 0) = Modelica.Constants.pi * diameter "Inner perimeter" annotation(
    //  Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = A_in * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    final parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 999.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 0.0011684 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    parameter Boolean friction = true "= true if there is friction in the system" annotation(
      Dialog(enable = true, tab = "General"));
    final parameter Real dx = length / moc.N "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    final parameter Integer T = moc.T;
    // Temporary time constant
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real v_array[T, N];
    //array for velocity
    Real fp_array[T, N];
    //array for friction at plus characteristic
    Real fm_array[T, N];
    //array for friction at minus characteristic
    Real Rp_array[T, N];
    //array for resistance at plus characteristic
    Real Rm_array[T, N];
    //array for resistance at minus characteristic
    Real Bp_array[T, N];
    //array for impedance at plus characteristic
    Real Bm_array[T, N];
    //array for impedance at minus characteristic
    Real Cp_array[T, N];
    //array for plus characteristic
    Real Cm_array[T, N];
    //array for minus characteristic
    Real H_array[T, N];
    //array for head
    Real Q_array[T, N];
    //array for flow
    Real A_vector[N];
    //vector for area
    Real D_vector[N];
    //vector for diameter
    // Initial condition declarations
    Real Qa = port_a.m_flow / rho;
    Real Qb = Qa;
    Real Ha = port_a.p / (rho * system.g);
    Real Hb = Ha;
    Real v = sqrt(2 * system.g * Ha);
    parameter Boolean expansion = false "= true if expansion equations are enabled. If not then A_outlet = A_inlet.";
  protected
  equation
//moc.time_step[1] = dx; not imlemented yet
/*
    if diameter_out == diameter_in then
      expansion = true;
    elseif not diameter_out == diameter_in then
      expansion = false;
    end if;
    
    Get translational warning: "In relation diameter_out == diameter_in, == on Real numbers is only allowed inside functions"
    Get an unbalanced system with one more equation than variables
    */
    A_vector[:] = linspace(A_in, A_out, N);
    D_vector[:] = linspace(diameter_in, diameter_out, N);
// Initial condition loop
    for k in 1:N loop
      H_array[1, k] = Ha;
      v_array[1, k] = OpenWPL.Functions.v_Velocity(H = H_array[1, k], g = system.g);
      Q_array[1, k] = v_array[1, k] * A_vector[k];
// Initial conditions for the plus characteristic
      if friction then
        fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[1, k] = 0.0;
      end if;
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rp_array[1, k] * abs(Q_array[1, k]);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = H_array[1, k], B = Bp_array[1, k], Qa = Q_array[1, k]);
// Initial conditions for the minus characteristic
      if friction then
        fm_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[1, k] = 0.0;
      end if;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rm_array[1, k] * abs(Q_array[1, k]);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = Hb, B = Bm_array[1, k], Qb = Q_array[1, k]);
    end for;
// End I.C loop
// Time iteration loop
    for j in 2:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
      H_array[j, 1] = Ha;
      v_array[j, 1] = OpenWPL.Functions.v_Velocity(H = H_array[j, 1], g = system.g);
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 2]) / Bm_array[j - 1, 2];
      if friction then
        fp_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, 1] = 0.0;
      end if;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = Bp_array[j, 1], Qa = Q_array[j, 1]);
      if friction then
        fm_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, 1] = 0.0;
      end if;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = Bm_array[j, 1], Qb = Q_array[j, 1]);
// BC right: closed by valve with Q=0
      if closedValve then
        Q_array[j, N] = 0.0;
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
      elseif not closedValve then
        Q_array[j, N] = Q_array[j, 1];
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
      end if;
      if friction then
        fp_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, N] = 0.0;
      end if;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = Bp_array[j, N], Qa = Q_array[j, N]);
      if friction then
        fm_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, N] = 0.0;
      end if;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = Bm_array[j, N], Qb = Q_array[j, N]);
// Space iteration loop
      for i in 2:N - 1 loop
        if expansion then
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_expansion(g = system.g, Ha = H_array[j - 1, i - 1], Hb = H_array[j - 1, i + 1], a = a, Aa = A_vector[i - 1], Ab = A_vector[i + 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Qb = Q_array[j - 1, i + 1])});
          H_array[j, i] = scalar({OpenWPL.Functions.H_expansion(g = system.g, Ha = H_array[j - 1, i - 1], a = a, Aa = A_vector[i - 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Q = Q_array[j, i])});
        elseif not expansion then
          H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        end if;
        v_array[j, i] = scalar(OpenWPL.Functions.v_Velocity(H = H_array[j, i], g = system.g));
        if friction then
          fp_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i - 1], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fp_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = D_vector[i - 1], A = A_vector[i - 1]);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i - 1], g = system.g) + Rp_array[j, i] * abs(Q_array[j - 1, i - 1]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = Bp_array[j, i], Qa = Q_array[j, i]);
        if friction then
          fm_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i + 1], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fm_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = D_vector[i + 1], A = A_vector[i + 1]);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i + 1], g = system.g) + Rm_array[j, i] * abs(Q_array[j - 1, i + 1]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = Bm_array[j, i], Qb = Q_array[j, i]);
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    0 = port_a.m_flow + port_b.m_flow;
//port_b.m_flow = Q_array[T,N]*rho;
//port_b.p = H_array[T,N]*rho*system.g;
// Port equations for substance mass flow. Not used, but implemented to be compatible with the Fluid library.
//port_a.Xi_outflow = inStream(port_b.Xi_outflow);
//port_b.Xi_outflow = inStream(port_a.Xi_outflow);
// Port equations for enthalpy flow. Not used, but implemented to be compatible with the Fluid library.
    port_b.h_outflow = inStream(port_a.h_outflow);
//- system.g*height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow);
//+ system.g*height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v9;

  model Pipe_MSL_v10
    /*
              Model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              Author:   Jonas Tveit Hinna
              Date:     30.04.2021
              Version:  10
            */
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    // Geometry
    final parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    final parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter_in "Inlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    parameter SI.Diameter diameter_out = diameter_in "Outlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area A_in = Modelica.Constants.pi * diameter_in * diameter_in / 4 "Inlet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Area A_out = Modelica.Constants.pi * diameter_out * diameter_out / 4 "Oulet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = A_in * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    final parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 999.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 0.0011684 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    parameter Boolean friction = true "= true if there is friction in the system" annotation(
      Dialog(enable = true, tab = "General"));
    final parameter Real dx = length / moc.N "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    final parameter Integer T = moc.T;
    // Temporary time constant
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real v_array[T, N];
    //array for velocity
    Real fp_array[T, N];
    //array for friction at plus characteristic
    Real fm_array[T, N];
    //array for friction at minus characteristic
    Real Rp_array[T, N];
    //array for resistance at plus characteristic
    Real Rm_array[T, N];
    //array for resistance at minus characteristic
    Real Bp_array[T, N];
    //array for impedance at plus characteristic
    Real Bm_array[T, N];
    //array for impedance at minus characteristic
    Real Cp_array[T, N];
    //array for plus characteristic
    Real Cm_array[T, N];
    //array for minus characteristic
    Real H_array[T, N];
    //array for head
    Real Q_array[T, N];
    //array for flow
    Real A_vector[N];
    //vector for area
    Real D_vector[N];
    //vector for diameter
    // Initial condition declarations
    Real Qa = port_a.m_flow / rho;
    Real Qb = Qa;
    Real Ha = port_a.p / (rho * system.g);
    Real Hb = Ha;
    Real v = sqrt(2 * system.g * Ha);
    parameter Boolean expansion = false "= true if expansion equations are enabled. If not then A_outlet = A_inlet.";
  protected
  equation
//moc.time_step[1] = dx; not imlemented yet
/*
    if diameter_out == diameter_in then
      expansion = true;
    elseif not diameter_out == diameter_in then
      expansion = false;
    end if;
    
    Get translational warning: "In relation diameter_out == diameter_in, == on Real numbers is only allowed inside functions"
    Get an unbalanced system with one more equation than variables
    */
    A_vector[:] = linspace(A_in, A_out, N);
    D_vector[:] = linspace(diameter_in, diameter_out, N);
// Initial condition loop
    for k in 1:N loop
      H_array[1, k] = Ha;
      v_array[1, k] = OpenWPL.Functions.v_Velocity(H = H_array[1, k], g = system.g);
      Q_array[1, k] = v_array[1, k] * A_vector[k];
// Initial conditions for the plus characteristic
      if friction then
        fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[1, k] = 0.0;
      end if;
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rp_array[1, k] * abs(Q_array[1, k]);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = H_array[1, k], B = Bp_array[1, k], Qa = Q_array[1, k]);
// Initial conditions for the minus characteristic
      if friction then
        fm_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[1, k] = 0.0;
      end if;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rm_array[1, k] * abs(Q_array[1, k]);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = H_array[1, k], B = Bm_array[1, k], Qb = Q_array[1, k]);
    end for;
// End I.C loop
// Time iteration loop
    for j in 2:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
      H_array[j, 1] = Ha;
      v_array[j, 1] = OpenWPL.Functions.v_Velocity(H = H_array[j, 1], g = system.g);
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 2]) / Bm_array[j - 1, 2];
      if friction then
        fp_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, 1] = 0.0;
      end if;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = Bp_array[j, 1], Qa = Q_array[j, 1]);
      if friction then
        fm_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, 1] = 0.0;
      end if;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = Bm_array[j, 1], Qb = Q_array[j, 1]);
// BC right: closed by valve with Q=0
      if closedValve then
        Q_array[j, N] = 0.0;
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
      elseif not closedValve then
        Q_array[j, N] = Q_array[j, 1];
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
      end if;
      if friction then
        fp_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, N] = 0.0;
      end if;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = Bp_array[j, N], Qa = Q_array[j, N]);
      if friction then
        fm_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, N] = 0.0;
      end if;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = Bm_array[j, N], Qb = Q_array[j, N]);
// Space iteration loop
      for i in 2:N - 1 loop
        if expansion then
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_expansion(g = system.g, Ha = H_array[j - 1, i - 1], Hb = H_array[j - 1, i + 1], a = a, Aa = A_vector[i - 1], Ab = A_vector[i + 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Qb = Q_array[j - 1, i + 1])});
          H_array[j, i] = scalar({OpenWPL.Functions.H_expansion(g = system.g, Ha = H_array[j - 1, i - 1], a = a, Aa = A_vector[i - 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Q = Q_array[j, i])});
        elseif not expansion then
          H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        end if;
        v_array[j, i] = scalar(OpenWPL.Functions.v_Velocity(H = H_array[j, i], g = system.g));
        if friction then
          fp_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fp_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rp_array[j, i] * abs(Q_array[j, i]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = Bp_array[j, i], Qa = Q_array[j, i]);
        if friction then
          fm_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fm_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rm_array[j, i] * abs(Q_array[j, i]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = Bm_array[j, i], Qb = Q_array[j, i]);
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    0 = port_a.m_flow + port_b.m_flow;
//port_b.m_flow = Q_array[T,N]*rho;
//port_b.p = H_array[T,N]*rho*system.g;
/*
    // Port equations for trace substance mass flow. Not used, but implemented to be compatible with the Fluid library.
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
    */
// Port equations for enthalpy flow. Not used, but implemented to be compatible with the Fluid library.
    port_b.h_outflow = inStream(port_a.h_outflow);
//- system.g*height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow);
//+ system.g*height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v10;

  model Pipe_MSL_v11
    /*
              Model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              Author:   Jonas Tveit Hinna
              Date:     18.05.2021
              Version:  11
            */
    import Modelica.Constants;
    outer Modelica.Fluid.System system "System wide properties";
    import Modelica.SIunits;
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    replaceable package Medium = Modelica.Media.Interfaces.PartialMedium "Medium in the component" annotation(
      choicesAllMatching = true);
    Modelica.Fluid.Interfaces.FluidPort_a port_a(redeclare package Medium = Medium, m_flow(min = if allowFlowReversal then -Constants.inf else 0)) "Fluid connector a (positive design flow direction is from port_a to port_b)" annotation(
      Placement(visible = true, transformation(extent = {{-110, -10}, {-90, 10}})));
    Modelica.Fluid.Interfaces.FluidPort_b port_b(redeclare package Medium = Medium, m_flow(max = if allowFlowReversal then +Constants.inf else 0)) "Fluid connector b (positive design flow direction is from port_a to port_b)" annotation(
      Placement(visible = true, transformation(extent = {{110, -10}, {90, 10}}), iconTransformation(extent = {{110, -10}, {90, 10}})));
    // Geometry
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    final parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter_in "Inlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    parameter SI.Diameter diameter_out = diameter_in "Outlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area A_in = Modelica.Constants.pi * diameter_in * diameter_in / 4 "Inlet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Area A_out = Modelica.Constants.pi * diameter_out * diameter_out / 4 "Oulet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = A_in * length "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    final parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 999.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 0.0011684 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    parameter Boolean friction = true "= true if there is friction in the system" annotation(
      Dialog(enable = true, tab = "General"));
    final parameter Real dx = length / moc.N "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    final parameter Integer T = moc.T;
    // Temporary time constant
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real v_array[T, N];
    //array for velocity
    Real fp_array[T, N];
    //array for friction at plus characteristic
    Real fm_array[T, N];
    //array for friction at minus characteristic
    Real Rp_array[T, N];
    //array for resistance at plus characteristic
    Real Rm_array[T, N];
    //array for resistance at minus characteristic
    Real Bp_array[T, N];
    //array for impedance at plus characteristic
    Real Bm_array[T, N];
    //array for impedance at minus characteristic
    Real Cp_array[T, N];
    //array for plus characteristic
    Real Cm_array[T, N];
    //array for minus characteristic
    Real H_array[T, N];
    //array for head
    Real Q_array[T, N];
    //array for flow
    Real A_vector[N];
    //vector for area
    Real D_vector[N];
    //vector for diameter
    // Initial condition declarations
    Real Qa = port_a.m_flow / rho;
    Real Qb = Qa;
    Real Ha = port_a.p / (rho * system.g);
    Real Hb = Ha;
    Real v = sqrt(2 * system.g * Ha);
    parameter Boolean expansion = false "= true if expansion equations are enabled. If not then A_outlet = A_inlet.";
  protected
  equation
//moc.time_step[1] = dx; not imlemented yet
/*
    if diameter_out == diameter_in then
      expansion = true;
    elseif not diameter_out == diameter_in then
      expansion = false;
    end if;
    
    Get translational warning: "In relation diameter_out == diameter_in, == on Real numbers is only allowed inside functions"
    Get an unbalanced system with one more equation than variables
    */
    A_vector[:] = linspace(A_in, A_out, N);
    D_vector[:] = linspace(diameter_in, diameter_out, N);
// Initial condition loop
    for k in 1:N loop
      H_array[1, k] = Ha;
      v_array[1, k] = OpenWPL.Functions.v_Velocity(H = H_array[1, k], g = system.g);
      Q_array[1, k] = v_array[1, k] * A_vector[k];
// Initial conditions for the plus characteristic
      if friction then
        fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[1, k] = 0.0;
      end if;
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rp_array[1, k] * abs(Q_array[1, k]);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = H_array[1, k], B = Bp_array[1, k], Qa = Q_array[1, k]);
// Initial conditions for the minus characteristic
      if friction then
        fm_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[1, k] = 0.0;
      end if;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rm_array[1, k] * abs(Q_array[1, k]);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = H_array[1, k], B = Bm_array[1, k], Qb = Q_array[1, k]);
    end for;
// End I.C loop
// Time iteration loop
    for j in 2:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
      H_array[j, 1] = Ha;
      v_array[j, 1] = OpenWPL.Functions.v_Velocity(H = H_array[j, 1], g = system.g);
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 2]) / Bm_array[j - 1, 2];
      if friction then
        fp_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, 1] = 0.0;
      end if;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = Bp_array[j, 1], Qa = Q_array[j, 1]);
      if friction then
        fm_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, 1] = 0.0;
      end if;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = Bm_array[j, 1], Qb = Q_array[j, 1]);
// BC right: closed by valve with Q=0
      if closedValve then
        Q_array[j, N] = 0.0;
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
      elseif not closedValve then
        Q_array[j, N] = Q_array[j, 1];
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
      end if;
      if friction then
        fp_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, N] = 0.0;
      end if;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = Bp_array[j, N], Qa = Q_array[j, N]);
      if friction then
        fm_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, N] = 0.0;
      end if;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = Bm_array[j, N], Qb = Q_array[j, N]);
// Space iteration loop
      for i in 2:N - 1 loop
        if expansion then
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_expansion(g = system.g, Ha = H_array[j - 1, i - 1], Hb = H_array[j - 1, i + 1], a = a, Aa = A_vector[i - 1], Ab = A_vector[i + 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Qb = Q_array[j - 1, i + 1])});
          H_array[j, i] = scalar({OpenWPL.Functions.H_expansion(g = system.g, Ha = H_array[j - 1, i - 1], a = a, Aa = A_vector[i - 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Q = Q_array[j, i])});
        elseif not expansion then
          H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        end if;
        v_array[j, i] = scalar(OpenWPL.Functions.v_Velocity(H = H_array[j, i], g = system.g));
        if friction then
          fp_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fp_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rp_array[j, i] * abs(Q_array[j, i]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = Bp_array[j, i], Qa = Q_array[j, i]);
        if friction then
          fm_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fm_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rm_array[j, i] * abs(Q_array[j, i]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = Bm_array[j, i], Qb = Q_array[j, i]);
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    0 = port_a.m_flow + port_b.m_flow;
//port_b.m_flow = Q_array[T,N]*rho;
//port_b.p = H_array[T,N]*rho*system.g;
/*
    // Port equations for trace substance mass flow. Not used, but implemented to be compatible with the Fluid library.
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
    */
// Port equations for enthalpy flow. Not used, but implemented to be compatible with the Fluid library.
    port_b.h_outflow = inStream(port_a.h_outflow);
//- system.g*height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow);
//+ system.g*height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      Icon(coordinateSystem(preserveAspectRatio = true, extent = {{-100, -100}, {100, 100}}), graphics = {Polygon(points = {{20, -70}, {60, -85}, {20, -100}, {20, -70}}, lineColor = {0, 128, 255}, fillColor = {0, 128, 255}, fillPattern = FillPattern.Solid, visible = showDesignFlowDirection), Polygon(points = {{20, -75}, {50, -85}, {20, -95}, {20, -75}}, lineColor = {255, 255, 255}, fillColor = {255, 255, 255}, fillPattern = FillPattern.Solid, visible = allowFlowReversal), Line(points = {{55, -85}, {-60, -85}}, color = {0, 128, 255}, visible = showDesignFlowDirection), Text(extent = {{-149, -114}, {151, -154}}, lineColor = {0, 0, 255}, textString = "%name"), Ellipse(extent = {{-110, 26}, {-90, -24}}, fillPattern = FillPattern.Solid, visible = port_a_exposesState), Ellipse(extent = {{90, 25}, {110, -25}}, fillPattern = FillPattern.Solid, visible = port_b_exposesState)}),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v11;

  model Pipe_MSL_v13
    /*
              Model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              Author:   Jonas Tveit Hinna
              Date:     21.05.2021
              Version:  13
            */
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    // Geometry
    final parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    final parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter_in "Inlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    parameter SI.Diameter diameter_out = diameter_in "Outlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area A_in = Modelica.Constants.pi * diameter_in * diameter_in / 4 "Inlet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Area A_out = Modelica.Constants.pi * diameter_out * diameter_out / 4 "Oulet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = A_in * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    final parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 999.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 0.0011684 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    parameter Boolean friction = true "= true if there is friction in the system" annotation(
      Dialog(enable = true, tab = "General"));
    final parameter Real dx = length / moc.N "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    final parameter Integer T = moc.T;
    // Temporary time constant
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real v_array[T, N];
    //array for velocity
    Real fp_array[T, N];
    //array for friction at plus characteristic
    Real fm_array[T, N];
    //array for friction at minus characteristic
    Real Rp_array[T, N];
    //array for resistance at plus characteristic
    Real Rm_array[T, N];
    //array for resistance at minus characteristic
    Real Bp_array[T, N];
    //array for impedance at plus characteristic
    Real Bm_array[T, N];
    //array for impedance at minus characteristic
    Real Cp_array[T, N];
    //array for plus characteristic
    Real Cm_array[T, N];
    //array for minus characteristic
    Real H_array[T, N];
    //array for head
    Real Q_array[T, N];
    //array for flow
    Real A_vector[N];
    //vector for area
    Real D_vector[N];
    //vector for diameter
    // Initial condition declarations
    Real Qa = port_a.m_flow / rho;
    Real Qb = Qa;
    Real Ha = port_a.p / (rho * system.g);
    //Real Hb = Ha;
    Real v = sqrt(2 * system.g * Ha);
    parameter Boolean expansion = false "= true if expansion equations are enabled. If not then A_outlet = A_inlet.";
  protected
  equation
//moc.time_step[1] = dx; not imlemented yet
/*
    if diameter_out == diameter_in then
      expansion = true;
    elseif not diameter_out == diameter_in then
      expansion = false;
    end if;
    
    Get translational warning: "In relation diameter_out == diameter_in, == on Real numbers is only allowed inside functions"
    Get an unbalanced system with one more equation than variables
    */
// Area and diameter vectors
    A_vector[:] = linspace(A_in, A_out, N);
    D_vector[:] = linspace(diameter_in, diameter_out, N);
// Initial condition loop
    for k in 1:N loop
      v_array[1, k] = OpenWPL.Functions.v_Velocity(H = Ha, g = system.g);
      Q_array[1, k] = v_array[1, k] * A_vector[k];
// Initial conditions for the plus characteristic
      if friction then
        fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[1, k] = 0.0;
      end if;
      H_array[1, k] = OpenWPL.Functions.H_initial(f = fp_array[1, k], H_res = Ha, Q = Q_array[1, k], L = dx * k, g = system.g, D = D_vector[k], A = A_vector[k]);
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rp_array[1, k] * abs(Q_array[1, k]);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = H_array[1, k], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g), Qa = Q_array[1, k]);
// Initial conditions for the minus characteristic
      if friction then
        fm_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[1, k] = 0.0;
      end if;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rm_array[1, k] * abs(Q_array[1, k]);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = H_array[1, k], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g), Qb = Q_array[1, k]);
    end for;
// End I.C loop
// Time iteration loop
    for j in 2:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
      H_array[j, 1] = Ha;
      v_array[j, 1] = OpenWPL.Functions.v_Velocity(H = H_array[j, 1], g = system.g);
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 2]) / Bm_array[j - 1, 2];
      if friction then
        fp_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, 1] = 0.0;
      end if;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g), Qa = Q_array[j, 1]);
      if friction then
        fm_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, 1] = 0.0;
      end if;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g), Qb = Q_array[j, 1]);
// BC right: closed by valve with Q=0
      if closedValve then
        Q_array[j, N] = 0.0;
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
      elseif not closedValve then
        Q_array[j, N] = Q_array[j, 1];
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
      end if;
      if friction then
        fp_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, N] = 0.0;
      end if;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g), Qa = Q_array[j, N]);
      if friction then
        fm_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, N] = 0.0;
      end if;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g), Qb = Q_array[j, N]);
// Space iteration loop
      for i in 2:N - 1 loop
        if expansion then
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_expansion(g = system.g, Ha = H_array[j - 1, i - 1], Hb = H_array[j - 1, i + 1], a = a, Aa = A_vector[i - 1], Ab = A_vector[i + 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Qb = Q_array[j - 1, i + 1])});
          H_array[j, i] = scalar({OpenWPL.Functions.H_expansion(g = system.g, Ha = H_array[j - 1, i - 1], a = a, Aa = A_vector[i - 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Q = Q_array[j, i])});
        elseif not expansion then
          H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        end if;
        v_array[j, i] = scalar(OpenWPL.Functions.v_Velocity(H = H_array[j, i], g = system.g));
        if friction then
          fp_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fp_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rp_array[j, i] * abs(Q_array[j, i]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g), Qa = Q_array[j, i]);
        if friction then
          fm_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fm_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rm_array[j, i] * abs(Q_array[j, i]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g), Qb = Q_array[j, i]);
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    0 = port_a.m_flow + port_b.m_flow;
//port_b.m_flow = Q_array[T,N]*rho;
//port_b.p = H_array[T,N]*rho*system.g;
/*
    // Port equations for trace substance mass flow. Not used, but implemented to be compatible with the Fluid library.
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
    */
// Port equations for enthalpy flow. Not used, but implemented to be compatible with the Fluid library.
    port_b.h_outflow = inStream(port_a.h_outflow);
//- system.g*height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow);
//+ system.g*height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v13;

  model Pipe_MSL_v14
    /*
              Model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              Author:   Jonas Tveit Hinna
              Date:     23.05.2021
              Version:  14
            */
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    // Geometry
    final parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    final parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter_in "Inlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    parameter SI.Diameter diameter_out = diameter_in "Outlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area A_in = Modelica.Constants.pi * diameter_in * diameter_in / 4 "Inlet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Area A_out = Modelica.Constants.pi * diameter_out * diameter_out / 4 "Oulet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = A_in * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    final parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 999.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 0.0011684 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    parameter Boolean friction = true "= true if there is friction in the system" annotation(
      Dialog(enable = true, tab = "General"));
    final parameter Real dx = length / moc.N "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    final parameter Integer T = moc.T;
    // Temporary time constant
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real v_array[T, N];
    //array for velocity
    Real fp_array[T, N];
    //array for friction at plus characteristic
    Real fm_array[T, N];
    //array for friction at minus characteristic
    Real Rp_array[T, N];
    //array for resistance at plus characteristic
    Real Rm_array[T, N];
    //array for resistance at minus characteristic
    Real Bp_array[T, N];
    //array for impedance at plus characteristic
    Real Bm_array[T, N];
    //array for impedance at minus characteristic
    Real Cp_array[T, N];
    //array for plus characteristic
    Real Cm_array[T, N];
    //array for minus characteristic
    Real H_array[T, N];
    //array for head
    Real Q_array[T, N];
    //array for flow
    Real A_vector[N];
    //vector for area
    Real D_vector[N];
    //vector for diameter
    // Initial condition declarations
    Real Qa = port_a.m_flow / rho;
    Real Qb = Qa;
    Real Ha = port_a.p / (rho * system.g);
    //Real Hb = Ha;
    Real v = sqrt(2 * system.g * Ha);
    parameter Boolean expansion = false "= true if expansion equations are enabled. If not then A_outlet = A_inlet.";
  protected
  equation
//moc.time_step[1] = dx; not imlemented yet
/*
    if diameter_out == diameter_in then
      expansion = true;
    elseif not diameter_out == diameter_in then
      expansion = false;
    end if;
    
    Get translational warning: "In relation diameter_out == diameter_in, == on Real numbers is only allowed inside functions"
    Get an unbalanced system with one more equation than variables
    */
// Area and diameter vectors
    A_vector[:] = linspace(A_in, A_out, N);
    D_vector[:] = linspace(diameter_in, diameter_out, N);
// Initial initial conditions
    H_array[1, 1] = Ha;
    v_array[1, 1] = OpenWPL.Functions.v_Velocity(H = H_array[1, 1], g = system.g);
    Q_array[1, 1] = v_array[1, 1] * A_vector[1];
    if friction then
      fp_array[1, 1] = OpenWPL.Functions.f_Haaland(v = v_array[1, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
    elseif not friction then
      fp_array[1, 1] = 0.0;
    end if;
    Rp_array[1, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[1, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
    Bp_array[1, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rp_array[1, 1] * abs(Q_array[1, 1]);
    Cp_array[1, 1] = OpenWPL.Functions.C_plus(Ha = H_array[1, 1], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g), Qa = Q_array[1, 1]);
    if friction then
      fm_array[1, 1] = OpenWPL.Functions.f_Haaland(v = v_array[1, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
    elseif not friction then
      fm_array[1, 1] = 0.0;
    end if;
    Rm_array[1, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[1, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
    Bm_array[1, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rm_array[1, 1] * abs(Q_array[1, 1]);
    Cm_array[1, 1] = OpenWPL.Functions.C_minus(Hb = H_array[1, 1], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g), Qb = Q_array[1, 1]);
// Initial condition loop
    for k in 2:N loop
      v_array[1, k] = OpenWPL.Functions.v_Velocity(H = H_array[1, k - 1], g = system.g);
      Q_array[1, k] = v_array[1, k] * A_vector[k];
// Initial conditions for the plus characteristic
      if friction then
        fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[1, k] = 0.0;
      end if;
      H_array[1, k] = OpenWPL.Functions.H_initial(f = fp_array[1, k], H_res = Ha, Q = Q_array[1, k], L = dx * k, g = system.g, D = D_vector[k], A = A_vector[k]);
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rp_array[1, k] * abs(Q_array[1, k]);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = H_array[1, k], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g), Qa = Q_array[1, k]);
// Initial conditions for the minus characteristic
      if friction then
        fm_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[1, k] = 0.0;
      end if;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rm_array[1, k] * abs(Q_array[1, k]);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = H_array[1, k], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g), Qb = Q_array[1, k]);
    end for;
// End I.C loop
// Time iteration loop
    for j in 2:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
      H_array[j, 1] = Ha;
      v_array[j, 1] = OpenWPL.Functions.v_Velocity(H = H_array[j, 1], g = system.g);
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 2]) / Bm_array[j - 1, 2];
      if friction then
        fp_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, 1] = 0.0;
      end if;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g), Qa = Q_array[j, 1]);
      if friction then
        fm_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, 1] = 0.0;
      end if;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g), Qb = Q_array[j, 1]);
// BC right: closed by valve with Q=0
      if closedValve then
        Q_array[j, N] = 0.0;
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
      elseif not closedValve then
        Q_array[j, N] = Q_array[j, 1];
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
      end if;
      if friction then
        fp_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, N] = 0.0;
      end if;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g), Qa = Q_array[j, N]);
      if friction then
        fm_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, N] = 0.0;
      end if;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g), Qb = Q_array[j, N]);
// Space iteration loop
      for i in 2:N - 1 loop
        if expansion then
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_expansion(g = system.g, Ha = H_array[j - 1, i - 1], Hb = H_array[j - 1, i + 1], a = a, Aa = A_vector[i - 1], Ab = A_vector[i + 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Qb = Q_array[j - 1, i + 1])});
          H_array[j, i] = scalar({OpenWPL.Functions.H_expansion(g = system.g, Ha = H_array[j - 1, i - 1], a = a, Aa = A_vector[i - 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Q = Q_array[j, i])});
        elseif not expansion then
          H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        end if;
        v_array[j, i] = scalar(OpenWPL.Functions.v_Velocity(H = H_array[j, i], g = system.g));
        if friction then
          fp_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fp_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rp_array[j, i] * abs(Q_array[j, i]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g), Qa = Q_array[j, i]);
        if friction then
          fm_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fm_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rm_array[j, i] * abs(Q_array[j, i]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g), Qb = Q_array[j, i]);
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    0 = port_a.m_flow + port_b.m_flow;
//port_b.m_flow = Q_array[T,N]*rho;
//port_b.p = H_array[T,N]*rho*system.g;
/*
    // Port equations for trace substance mass flow. Not used, but implemented to be compatible with the Fluid library.
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
    */
// Port equations for enthalpy flow. Not used, but implemented to be compatible with the Fluid library.
    port_b.h_outflow = inStream(port_a.h_outflow);
//- system.g*height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow);
//+ system.g*height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v14;

  model Pipe_MSL_v15
    /*
              Model of a pipe using MOC to simulate the waterhammer effect and mass oscillations.
              
              Author:   Jonas Tveit Hinna
              Date:     31.05.2021
              Version:  15
            */
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    // Geometry
    final parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    final parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter_in "Inlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    parameter SI.Diameter diameter_out = diameter_in "Outlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area A_in = Modelica.Constants.pi * diameter_in * diameter_in / 4 "Inlet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Area A_out = Modelica.Constants.pi * diameter_out * diameter_out / 4 "Oulet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = A_in * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    final parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 999.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 0.0011684 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    parameter Boolean friction = true "= true if there is friction in the system" annotation(
      Dialog(enable = true, tab = "General"));
    final parameter Real dx = length / moc.N "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    final parameter Integer T = moc.T;
    // Temporary time constant
    final parameter Integer H_reservoir = moc.H_reservoir;
    final parameter Integer H_sink = moc.H_sink;
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real v_array[T, N];
    //array for velocity
    Real fp_array[T, N];
    //array for friction at plus characteristic
    Real fm_array[T, N];
    //array for friction at minus characteristic
    Real Rp_array[T, N];
    //array for resistance at plus characteristic
    Real Rm_array[T, N];
    //array for resistance at minus characteristic
    Real Bp_array[T, N];
    //array for impedance at plus characteristic
    Real Bm_array[T, N];
    //array for impedance at minus characteristic
    Real Cp_array[T, N];
    //array for plus characteristic
    Real Cm_array[T, N];
    //array for minus characteristic
    Real H_array[T, N];
    //array for head
    Real Q_array[T, N];
    //array for flow
    Real A_vector[N];
    //vector for area
    Real D_vector[N];
    //vector for diameter
    // Initial condition declarations
    Real Qa = port_a.m_flow / rho;
    Real Qb = Qa;
    Real Ha = port_a.p / (rho * system.g);
    //Real Hb = Ha;
    Real v = sqrt(2 * system.g * Ha);
    Real hf_guess;
    Real v_guess;
    Real v_guessed;
    Real f_guess;
    //Real v_iterate;
    Real dh;
    parameter Boolean expansion = false "= true if expansion equations are enabled. If not then A_outlet = A_inlet.";
  algorithm
// Area and diameter vectors
    A_vector[:] := linspace(A_in, A_out, N);
    D_vector[:] := linspace(diameter_in, diameter_out, N);
    v_guess := OpenWPL.Functions.v_Velocity(H = Ha, g = system.g);
    f_guess := OpenWPL.Functions.f_Haaland(v = v_guess, D = D_vector[1], rho = rho, mu = mu, eps = roughness);
    hf_guess := OpenWPL.Functions.H_f(f = f_guess, L = length, D = D_vector[1], V = v_guess, g = system.g);
    dh := H_reservoir - H_sink;
    while hf_guess >= dh loop
      v_guess := v_guess * 0.95;
      f_guess := 100 * OpenWPL.Functions.f_Haaland(v = v_guess, D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      hf_guess := OpenWPL.Functions.H_f(f = f_guess, L = length, D = D_vector[1], V = v_guess, g = system.g);
    end while;
    v_guessed := v_guess;
  equation
//moc.time_step[1] = dx; not imlemented yet
/*
    if diameter_out == diameter_in then
      expansion = true;
    elseif not diameter_out == diameter_in then
      expansion = false;
    end if;
    
    Get translational warning: "In relation diameter_out == diameter_in, == on Real numbers is only allowed inside functions"
    Get an unbalanced system with one more equation than variables
    */
//v_guess = OpenWPL.Functions.v_Velocity(H=Ha, g = system.g);
//f_guess = OpenWPL.Functions.f_Haaland(v = v_guess, D = D_vector[1], rho = rho, mu = mu, eps = roughness);
// Initial condition loop
    for k in 1:N loop
//v_array[1, k] = OpenWPL.Functions.v_Velocity(H=Ha, g = system.g);
//fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
//hf = OpenWPL.Functions.H_f(f=fp_array[1,k],L=length,D=D_vector[k],V=v_array[1,k],g=system.g);
//v_array[1,k] = OpenWPL.Functions.v_iterate(D=D_vector[k], rho=rho, mu=mu, eps=roughness , L=length, dh=(H_reservoir-H_sink), v_guess=v,hf_guess=OpenWPL.Functions.H_f(f=OpenWPL.Functions.f_Haaland(v = v, D = D_vector[k], rho = rho, mu = mu, eps = roughness),L=length,D=D_vector[k],V=OpenWPL.Functions.v_Velocity(H=Ha, g = system.g),g=system.g), g=system.g);
/*
      while hf<=(H_reservoir-H_sink) loop
        v_array[1,k] = v_array[1,k]*0.95;  
        fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
        hf = OpenWPL.Functions.H_f(f=fp_array[1,k],L=length,D=D_vector[k],V=v_array[1,k],g=system.g);
      end while;
      */
// Initial conditions for the plus characteristic
      if friction then
        fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[1, k] = 0.0;
      end if;
//v_array[1,k] = OpenWPL.Functions.v_Bernoulli(H_in=H_reservoir,H_out=H_sink,g = system.g,k=0,f=fp_array[1,k],L=length,D=D_vector[k]);
      v_array[1, k] = v_guessed;
      Q_array[1, k] = v_array[1, k] * A_vector[k];
      H_array[1, k] = OpenWPL.Functions.H_initial(f = fp_array[1, k], H_res = Ha, Q = Q_array[1, k], L = dx * k, g = system.g, D = D_vector[k], A = A_vector[k]);
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rp_array[1, k] * abs(Q_array[1, k]);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = H_array[1, k], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g), Qa = Q_array[1, k]);
// Initial conditions for the minus characteristic
      if friction then
        fm_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[1, k] = 0.0;
      end if;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rm_array[1, k] * abs(Q_array[1, k]);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = H_array[1, k], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g), Qb = Q_array[1, k]);
    end for;
// End I.C loop
// Time iteration loop
    for j in 2:T loop
// BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
      H_array[j, 1] = Ha;
//v_array[j, 1] = OpenWPL.Functions.v_Velocity(H = H_array[j, 1], g = system.g);
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 2]) / Bm_array[j - 1, 2];
      v_array[j, 1] = Q_array[j, 1] / A_vector[1];
      if friction then
        fp_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, 1] = 0.0;
      end if;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g), Qa = Q_array[j, 1]);
      if friction then
        fm_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, 1] = 0.0;
      end if;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g), Qb = Q_array[j, 1]);
// BC right: closed by valve with Q=0
      if closedValve then
        Q_array[j, N] = 0.0;
        v_array[j, N] = 0.0;
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
      elseif not closedValve then
        Q_array[j, N] = Q_array[j, 1];
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
        v_array[j, N] = Q_array[j, N] / A_vector[N];
//removed v_velocity function here
      end if;
      if friction then
        fp_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, N] = 0.0;
      end if;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g), Qa = Q_array[j, N]);
      if friction then
        fm_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, N] = 0.0;
      end if;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g), Qb = Q_array[j, N]);
// Space iteration loop
      for i in 2:N - 1 loop
        if expansion then
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_expansion(g = system.g, Ha = H_array[j - 1, i - 1], Hb = H_array[j - 1, i + 1], a = a, Aa = A_vector[i - 1], Ab = A_vector[i + 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Qb = Q_array[j - 1, i + 1])});
          H_array[j, i] = scalar({OpenWPL.Functions.H_expansion(g = system.g, Ha = H_array[j - 1, i - 1], a = a, Aa = A_vector[i - 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Q = Q_array[j, i])});
        elseif not expansion then
          H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        end if;
//v_array[j, i] = scalar(OpenWPL.Functions.v_Velocity(H = H_array[j, i], g = system.g));
        v_array[j, i] = Q_array[j, i] / A_vector[i];
        if friction then
          fp_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fp_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rp_array[j, i] * abs(Q_array[j, i]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g), Qa = Q_array[j, i]);
        if friction then
          fm_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fm_array[j, i] = 0.0;
// if testing for a square pressure pulse
        end if;
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rm_array[j, i] * abs(Q_array[j, i]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g), Qb = Q_array[j, i]);
      end for;
// End space loop
    end for;
// End time loop
// Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    0 = port_a.m_flow + port_b.m_flow;
//port_b.m_flow = Q_array[T,N]*rho;
//port_b.p = H_array[T,N]*rho*system.g;
/*
    // Port equations for trace substance mass flow. Not used, but implemented to be compatible with the Fluid library.
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
    */
// Port equations for enthalpy flow. Not used, but implemented to be compatible with the Fluid library.
    port_b.h_outflow = inStream(port_a.h_outflow);
//- system.g*height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow);
//+ system.g*height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics.
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe_MSL_v15;

  model Pipe
    /*
              Final model of a pipe using MOC to simulate the waterhammer effect and mass oscillations. Based on version 13.
              
              Author:   Jonas Tveit Hinna
              Date:     10.06.2021
              Version:  13
            */
    extends Modelica.Fluid.Interfaces.PartialTwoPort;
    import Modelica.Constants;
    import Modelica.SIunits;
    outer Modelica.Fluid.System system "System wide properties";
    outer OpenWPL.MOC moc "System wide storage of MOC values";
    extends OpenWPL.Icons.Waterhammer_MOC;
    // Geometry
    final parameter Real nParallel(min = 1) = 1 "Number of identical parallel pipes" annotation(
      Dialog(group = "Geometry"));
    parameter SI.Length length = 1 "Length" annotation(
      Dialog(tab = "General", group = "Geometry"));
    final parameter Boolean isCircular = true "= true if cross sectional area is circular" annotation(
      Evaluate,
      Dialog(tab = "General", group = "Geometry"));
    parameter SI.Diameter diameter_in "Inlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    parameter SI.Diameter diameter_out = diameter_in "Outlet diameter of circular pipe" annotation(
      Dialog(group = "Geometry", enable = isCircular));
    final parameter SI.Area A_in = Modelica.Constants.pi * diameter_in * diameter_in / 4 "Inlet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter SI.Area A_out = Modelica.Constants.pi * diameter_out * diameter_out / 4 "Oulet cross section area" annotation(
      Dialog(tab = "General", group = "Geometry", enable = not isCircular));
    final parameter Modelica.Fluid.Types.Roughness roughness = 2.5e-5 "Average height of surface asperities" annotation(
      Dialog(group = "Geometry"));
    final parameter SI.Volume V = A_in * length * nParallel "volume size";
    // Boundary condition determination
    parameter Boolean closedValve = true "= true if the right side of the pipe is closed by a valve" annotation(
      Dialog(enable = true, tab = "General", group = "Boundary condition"));
    // Static head
    final parameter SI.Length height_ab = 0 "Height(port_b) - Height(port_a)" annotation(
      Dialog(group = "Static head"));
    // Initialization
    final parameter Medium.AbsolutePressure p_a_start = system.p_start "Start value of pressure at port a" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.AbsolutePressure p_b_start = p_a_start "Start value of pressure at port b" annotation(
      Dialog(tab = "Initialization"));
    final parameter Medium.MassFlowRate m_flow_start = system.m_flow_start "Start value for mass flow rate" annotation(
      Evaluate = true,
      Dialog(tab = "Initialization"));
    // Fluid parameters
    final parameter SIunits.Density rho = 999.2 "Fluid density" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.DynamicViscosity mu = 0.0011684 "Dynamic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    final parameter SIunits.KinematicViscosity nu = mu / rho "Kinematic viscosity" annotation(
      Dialog(group = "Fluid properties"));
    parameter SIunits.VelocityOfSound a = 1200 "Velocity of sound in water" annotation(
      Dialog(group = "Fluid properties"));
    // Method of Characteristics declarations
    parameter Boolean friction = true "= true if there is friction in the system" annotation(
      Dialog(enable = true, tab = "General"));
    final parameter Real dx = length / moc.N "Length step calculated in this model";
    final parameter Integer N = integer(length / dx) "n parts";
    final parameter Real dt = dx / a "Time step";
    //final parameter Real dx_min = moc.dx_min;
    final parameter Real dx_min = dx;
    final parameter Integer T = moc.T;
    // Temporary time constant
    // Local array storage declarations. Global values are stored in the component "MOC.mo"
    Real v_array[T, N];
    //array for velocity
    Real fp_array[T, N];
    //array for friction at plus characteristic
    Real fm_array[T, N];
    //array for friction at minus characteristic
    Real Rp_array[T, N];
    //array for resistance at plus characteristic
    Real Rm_array[T, N];
    //array for resistance at minus characteristic
    Real Bp_array[T, N];
    //array for impedance at plus characteristic
    Real Bm_array[T, N];
    //array for impedance at minus characteristic
    Real Cp_array[T, N];
    //array for plus characteristic
    Real Cm_array[T, N];
    //array for minus characteristic
    Real H_array[T, N];
    //array for head
    Real Q_array[T, N];
    //array for flow
    Real A_vector[N];
    //vector for area
    Real D_vector[N];
    //vector for diameter
    // Initial condition declarations
    Real Qa = port_a.m_flow / rho;
    Real Qb = Qa;
    Real Ha = port_a.p / (rho * system.g);
    //Real Hb = Ha;
    Real v = sqrt(2 * system.g * Ha);
    parameter Boolean expansion = false "= true if expansion equations are enabled. If not then A_outlet = A_inlet.";
  protected
  equation
  //moc.time_step[1] = dx; not imlemented yet
  /*
    if diameter_out == diameter_in then
      expansion = true;
    elseif not diameter_out == diameter_in then
      expansion = false;
    end if;
    
    Get translational warning: "In relation diameter_out == diameter_in, == on Real numbers is only allowed inside functions"
    Get an unbalanced system with one more equation than variables
    */
  // Area and diameter vectors
    A_vector[:] = linspace(A_in, A_out, N);
    D_vector[:] = linspace(diameter_in, diameter_out, N);
  // Initial condition loop
    for k in 1:N loop
      v_array[1, k] = OpenWPL.Functions.v_Velocity(H = Ha, g = system.g);
      Q_array[1, k] = v_array[1, k] * A_vector[k];
  // Initial conditions for the plus characteristic
      if friction then
        fp_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[1, k] = 0.0;
      end if;
      H_array[1, k] = OpenWPL.Functions.H_initial(f = fp_array[1, k], H_res = Ha, Q = Q_array[1, k], L = dx * k, g = system.g, D = D_vector[k], A = A_vector[k]);
      Rp_array[1, k] = OpenWPL.Functions.R_Resistance(f = fp_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bp_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rp_array[1, k] * abs(Q_array[1, k]);
      Cp_array[1, k] = OpenWPL.Functions.C_plus(Ha = H_array[1, k], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g), Qa = Q_array[1, k]);
  // Initial conditions for the minus characteristic
      if friction then
        fm_array[1, k] = OpenWPL.Functions.f_Haaland(v = v_array[1, k], D = D_vector[k], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[1, k] = 0.0;
      end if;
      Rm_array[1, k] = OpenWPL.Functions.R_Resistance(f = fm_array[1, k], dx = dx_min, g = system.g, D = D_vector[k], A = A_vector[k]);
      Bm_array[1, k] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g) + Rm_array[1, k] * abs(Q_array[1, k]);
      Cm_array[1, k] = OpenWPL.Functions.C_minus(Hb = H_array[1, k], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[k], g = system.g), Qb = Q_array[1, k]);
    end for;
  // End I.C loop
  // Time iteration loop
    for j in 2:T loop
  // BC left: open to reservoir with p equal to port.p (should be dp/dt=0)
      H_array[j, 1] = Ha;
      v_array[j, 1] = OpenWPL.Functions.v_Velocity(H = H_array[j, 1], g = system.g);
      Q_array[j, 1] = (H_array[j, 1] - Cm_array[j - 1, 2]) / Bm_array[j - 1, 2];
      if friction then
        fp_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, 1] = 0.0;
      end if;
      Rp_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fp_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bp_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rp_array[j, 1] * abs(Q_array[j, 1]);
      Cp_array[j, 1] = OpenWPL.Functions.C_plus(Ha = H_array[j, 1], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g), Qa = Q_array[j, 1]);
      if friction then
        fm_array[j, 1] = OpenWPL.Functions.f_Haaland(v = v_array[j, 1], D = D_vector[1], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, 1] = 0.0;
      end if;
      Rm_array[j, 1] = OpenWPL.Functions.R_Resistance(f = fm_array[j, 1], dx = dx_min, g = system.g, D = D_vector[1], A = A_vector[1]);
      Bm_array[j, 1] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g) + Rm_array[j, 1] * abs(Q_array[j, 1]);
      Cm_array[j, 1] = OpenWPL.Functions.C_minus(Hb = H_array[j, 1], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[1], g = system.g), Qb = Q_array[j, 1]);
  // BC right: closed by valve with Q=0
      if closedValve then
        Q_array[j, N] = 0.0;
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
      elseif not closedValve then
        Q_array[j, N] = Q_array[j, 1];
        H_array[j, N] = Cp_array[j - 1, N - 1] - Bp_array[j - 1, N - 1] * Q_array[j, N];
        v_array[j, N] = OpenWPL.Functions.v_Velocity(H = H_array[j, N], g = system.g);
      end if;
      if friction then
        fp_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fp_array[j, N] = 0.0;
      end if;
      Rp_array[j, N] = OpenWPL.Functions.R_Resistance(f = fp_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bp_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rp_array[j, N] * abs(Q_array[j, N]);
      Cp_array[j, N] = OpenWPL.Functions.C_plus(Ha = H_array[j, N], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g), Qa = Q_array[j, N]);
      if friction then
        fm_array[j, N] = OpenWPL.Functions.f_Haaland(v = v_array[j, N], D = D_vector[N], rho = rho, mu = mu, eps = roughness);
      elseif not friction then
        fm_array[j, N] = 0.0;
      end if;
      Rm_array[j, N] = OpenWPL.Functions.R_Resistance(f = fm_array[j, N], dx = dx_min, g = system.g, D = D_vector[N], A = A_vector[N]);
      Bm_array[j, N] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g) + Rm_array[j, N] * abs(Q_array[j, N]);
      Cm_array[j, N] = OpenWPL.Functions.C_minus(Hb = H_array[j, N], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[N], g = system.g), Qb = Q_array[j, N]);
  // Space iteration loop
      for i in 2:N - 1 loop
        if expansion then
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_expansion(g = system.g, Ha = H_array[j - 1, i - 1], Hb = H_array[j - 1, i + 1], a = a, Aa = A_vector[i - 1], Ab = A_vector[i + 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Qb = Q_array[j - 1, i + 1])});
          H_array[j, i] = scalar({OpenWPL.Functions.H_expansion(g = system.g, Ha = H_array[j - 1, i - 1], a = a, Aa = A_vector[i - 1], A = A_vector[i], Qa = Q_array[j - 1, i - 1], Q = Q_array[j, i])});
        elseif not expansion then
          H_array[j, i] = scalar({OpenWPL.Functions.H_head(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
          Q_array[j, i] = scalar({OpenWPL.Functions.Q_flow(Cp = Cp_array[j - 1, i - 1], Cm = Cm_array[j - 1, i + 1], Bm = Bm_array[j - 1, i + 1], Bp = Bp_array[j - 1, i - 1])});
        end if;
        v_array[j, i] = scalar(OpenWPL.Functions.v_Velocity(H = H_array[j, i], g = system.g));
        if friction then
          fp_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fp_array[j, i] = 0.0;
  // if testing for a square pressure pulse
        end if;
        Rp_array[j, i] = OpenWPL.Functions.R_Resistance(f = fp_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bp_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rp_array[j, i] * abs(Q_array[j, i]);
        Cp_array[j, i] = OpenWPL.Functions.C_plus(Ha = H_array[j, i], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g), Qa = Q_array[j, i]);
        if friction then
          fm_array[j, i] = OpenWPL.Functions.f_Haaland(v = v_array[j, i], D = D_vector[i], rho = rho, mu = mu, eps = roughness);
        elseif not friction then
          fm_array[j, i] = 0.0;
  // if testing for a square pressure pulse
        end if;
        Rm_array[j, i] = OpenWPL.Functions.R_Resistance(f = fm_array[j, i], dx = dx_min, g = system.g, D = D_vector[i], A = A_vector[i]);
        Bm_array[j, i] = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g) + Rm_array[j, i] * abs(Q_array[j, i]);
        Cm_array[j, i] = OpenWPL.Functions.C_minus(Hb = H_array[j, i], B = OpenWPL.Functions.B_Impedance(a = a, A = A_vector[i], g = system.g), Qb = Q_array[j, i]);
      end for;
  // End space loop
    end for;
  // End time loop
  // Port equations for mass flow rate
    port_a.m_flow = system.m_flow_start;
    0 = port_a.m_flow + port_b.m_flow;
  //port_b.m_flow = Q_array[T,N]*rho;
  //port_b.p = H_array[T,N]*rho*system.g;
  /*
    // Port equations for trace substance mass flow. Not used, but implemented to be compatible with the Fluid library.
    port_a.C_outflow = inStream(port_b.C_outflow);
    port_b.C_outflow = inStream(port_a.C_outflow);
    */
  // Port equations for enthalpy flow. Not used, but implemented to be compatible with the Fluid library.
    port_b.h_outflow = inStream(port_a.h_outflow);
  //- system.g*height_ab;
    port_a.h_outflow = inStream(port_b.h_outflow);
  //+ system.g*height_ab;
    annotation(
      defaultComponentName = "pipe",
      Documentation(info = "<html>
  <p>Model of a straight pipe modelled using the method of characteristics. This is the newest version of the component (same as v13).
  </p>
  </html>"),
      __OpenModelica_commandLineOptions = "--matchingAlgorithm=PFPlusExt --indexReductionMethod=dynamicStateSelection -d=initialization,NLSanalyticJacobian,newInst");
  end Pipe;
  annotation(
    Icon(graphics = {Polygon(points = {{-70, 26}, {68, -44}, {68, 26}, {2, -10}, {-70, -42}, {-70, 26}}), Line(points = {{2, 42}, {2, -10}}), Rectangle(extent = {{-18, 50}, {22, 42}}, fillPattern = FillPattern.Solid)}),
    preferredView = "info",
    Documentation(info = "<html>
<p>
These components are made as a results of a master thesis conducted at the Waterpower Laboratory at NTNU by Jonas Tveit Hinna in 2021.
</html>"));
end Components;
